from dash import Dash, Output, Input, State, html, dcc, callback, MATCH,clientside_callback,dash_table,ALL,no_update,ctx
import uuid
import dash_bootstrap_components as dbc
from src._internal import Author
from typing import Iterable,Union,Callable,Optional
from math import ceil
from src.utility.filters import FilterFactory,Filter,sorter_provider
import src.gui.filters
import json
import pandas as pd
import re
from logging import getLogger

logger = getLogger("components_logger")

class CustomTable():
    class ids:
        modal = lambda aio_id: {
            'component': 'CustomTable',
            'subcomponent': 'modal',
            'aio_id': aio_id
        }
        button = lambda aio_id,target,purpose: {
            'component': 'CustomTable',
            'subcomponent': 'button',
            'aio_id': aio_id,
            'target': target,
            'purpose': purpose
        }
        pagination =lambda aio_id: {
            'component': 'CustomTable',
            'subcomponent': 'pagination',
            'aio_id': aio_id,
        }
        table =lambda aio_id: {
            'component': 'CustomTable',
            'subcomponent': 'table',
            'aio_id': aio_id
        }
        store =lambda aio_id,sub_id: {
            'component': 'CustomTable',
            'subcomponent': 'store',
            'aio_id': aio_id,
            'sub_id':sub_id
        }
        input= lambda aio_id,sub_id,target: {
            'component': 'CustomTable',
            'subcomponent': 'input',
            'aio_id': aio_id,
            'sub_id':sub_id,
            'target':target
        }
        
    def __init__(
        self,
        data:Union[dict[str,list]|list[dict[str]]],
        filters:dict[str,str]=None,
        sort:dict[str,str]=None,
        elements_per_page:int=5,
        div_props:dict=None,
        table_props:dict=None,
        id:str=None,
    ):
        data_to_use:list[dict[str]]=list()
        keys=None
        ln=0
        if isinstance(data,dict):
            keys=sorted(list(data.keys()))
            key=keys.pop()
            keys.add(key)
            ln=len(data[key])
            #check for equal length in data lists
            for k,v in data.items():
                if not isinstance(v,list):
                    raise TypeError("The shape of dictionary must be dict[str,list] or list[dict[str]]")

            for i in range(ln):
                nd=dict()
                for k in data.keys():
                    try:
                        nd[k]=data[k][i]
                    except IndexError:
                        raise ValueError("All data must have the same length")
                data_to_use.append(nd)
                
        elif isinstance(data,list):
            ln=len(data)
            if not data:
                raise ValueError("Cannot compute an empty list")
            keys=sorted(data[0].keys())
            for d in data:
                if not isinstance(d,dict):
                    raise TypeError("The shape of dictionary must be dict[str,list] or list[dict[str]]")
                if set(d.keys()) != set(keys):
                    raise ValueError("All data must have the same keys")
            data_to_use=data
        
        num_pages=ceil(ln/elements_per_page)
        if id is None:
            id = str(uuid.uuid4())
        d_props =div_props.copy() if div_props else {}
        t_props =table_props.copy() if table_props else {}
        
        
        table_header = [html.Thead(html.Tr([html.Th([
            html.Span(" ".join(header.split("_")).capitalize(),className="pe-2"),
            html.I(className="bi bi-funnel clickable",id=self.ids.button(id,header,"filter"),n_clicks=0) if filters and header in filters else "",
            dbc.Collapse([
                dbc.Row(
                    [
                        dbc.Col([
                            dbc.Input(id=self.ids.input(id,"text",header),placeholder="Filter",size="md")
                            ],width=10),
                        dbc.Col([
                            html.I(className="bi bi-x-lg clickable",id=self.ids.button(id,header,"filter_clear"),n_clicks=0)
                            ],width=2,align="center")
                        ]),
                ],id=self.ids.input(id,"collapse",header),is_open=False) if filters and header in filters else "",
            html.I(className="bi bi-caret-down clickable",id=self.ids.button(id,header,"sort"),n_clicks=0) if sort and  header in sort else "",
            ]) for header in keys]))]
        
        table = dbc.Table(table_header,id=self.ids.table(id),**t_props)
        self.comp=dbc.Container([
            dcc.Store(id=self.ids.store(id,"filters"),data={"filters":filters,"sorters":sort}),
            dcc.Store(id=self.ids.store(id,"filtered_indexes"),data=list(range(len(data_to_use)))),
            dcc.Store(id=self.ids.store(id,"full_data"),data=data_to_use),
            dcc.Store(id=self.ids.store(id,"pagination"),data=dict(epg=elements_per_page)),
            dbc.Pagination(id=self.ids.pagination(id),min_value=1,max_value=num_pages,active_page=1,first_last=True,previous_next=True,fully_expanded=False),
            table
        ],**d_props)

        
    def create_comp(self):
        return self.comp

    @callback(
        Output(ids.table(MATCH), 'children'),
        Output(ids.pagination(MATCH), 'max_value'),
        Input(ids.pagination(MATCH), 'active_page'),
        Input(ids.store(MATCH,"filtered_indexes"),"data"),
        Input(ids.store(MATCH,"full_data"),"data"),
        State(ids.store(MATCH,"pagination"), 'data'),
        State(ids.table(MATCH), 'children'),
    )
    def populate_table(page,filtered_indexes:list[int],data,pag_data,table_data):
        # print(table_data[0])
        if not (pag_data and data):
            return no_update
        data_to_show=list()
        logger.debug(f"Filter indexes for population",extra={"component":"CustomTable","data":filtered_indexes})
        for i in filtered_indexes:
            data_to_show.append(data[i])
        epg=pag_data["epg"]
        index=epg*(page-1)
        rows:list[html.Tr]=list()
        if data_to_show:
            keys=sorted(data_to_show[0].keys())
            for d in data_to_show[index:index+epg]:
                rows.append(html.Tr([html.Td(d[k]) for k in keys]))
        else:
            rows.append("no data")
        table_body = [table_data[0],html.Tbody(rows)]
        return table_body,ceil(len(data_to_show)/epg)
    

    clientside_callback(
        """
        function(_){
            if(_== undefined){
                return window.dash_clientside.no_update;
            }
            const is_open=_%2==1;
            var collapse_control=  is_open ? ['bi bi-funnel-fill clickable',is_open] : ['bi bi-funnel clickable',is_open];
            return collapse_control;
            }   
        """,
        Output(ids.button(MATCH,MATCH,"filter"), 'className'),
        Output(ids.input(MATCH,"collapse",MATCH), 'is_open'),
        Input(ids.button(MATCH,MATCH,"filter"), 'n_clicks'),
    )
    
    @callback(
        Output(ids.store(MATCH,"filtered_indexes"),"data"),
        Input(ids.input(MATCH,"text",ALL), 'value'),
        State(ids.store(MATCH,"full_data"),"data"),
        State(ids.store(MATCH,"filters"), 'data'),
        State(ids.input(MATCH,"text",ALL), 'id'),
        prevent_initial_call=True
    )
    def filter_data(text,orig_data,filters,in_id):
        print(ctx.triggered_id)
        if not in_id[0] or not text:
            return no_update
        filtered_indexes=set()
        text_dict=dict()
        filter_dict:dict[str,Filter]=dict()
        for t,id_dict in zip(text,in_id):
            text_dict[id_dict["target"]]=t
            filter_dict[id_dict["target"]]=FilterFactory.create_filter(filters["filters"][id_dict["target"]])
        for i,d in enumerate(orig_data):
            for target,text_to_use in text_dict.items():
                filter=filter_dict[target]
                res=filter.fun(text_to_use,d[target])
                # logger.debug(f"Filter result for {target} is {res} ",extra={"component":"CustomTable"})
                if res:
                    filtered_indexes.add(i)
                else:
                    filtered_indexes.discard(i)
                    break
        filtered_indexes=list(filtered_indexes)
        logger.debug(f"Filter indexes from filter callback",extra={"component":"CustomTable","data":filtered_indexes})
        return filtered_indexes
    
    clientside_callback(
        """
        function(_){
            if(_== undefined){
                return window.dash_clientside.no_update;
            }
            const up=_%2==1;
            var collapse_control=  up ? 'bi bi-caret-up clickable' : 'bi bi-caret-down clickable';
            return collapse_control;
            }   
        """,
        Output(ids.button(MATCH,MATCH,"sort"), 'className'),
        Input(ids.button(MATCH,MATCH,"sort"), 'n_clicks'),
    )
    
    clientside_callback(
        """
        function(_){
            if(_== undefined){
                return window.dash_clientside.no_update;
            }
            return "";
            }   
        """,
        Output(ids.input(MATCH,"text",MATCH), 'value'),
        Input(ids.button(MATCH,MATCH,"filter_clear"), 'n_clicks'),
        prevent_initial_call=True
    )

    @callback(
        Output(ids.store(MATCH,"full_data"),"data"),
        Input(ids.button(MATCH,ALL,"sort"),"n_clicks"),
        State(ids.store(MATCH,"full_data"),"data"),
        State(ids.store(MATCH,"filters"), 'data'),
        prevent_initial_call=True
    )
    def sort_data(_,data,filters):
        if data==None or _==None or _==0:
            return no_update
        target=ctx.triggered_id["target"]
        # ids=btn_id
        logger.debug(f"Sort: {target}",extra={"component":"CustomTable","data":data[0][target]})
        sorter = sorter_provider(filters["sorters"][target])
        reverse=_[0]%2==0
        logger.debug(f"Chose sorter: {filters['sorters'][target]}",extra={"component":"CustomTable"})
        if not sorter:
            sorted_data=sorted(data,key=lambda d: d[target],reverse=reverse)
        else:
            sorted_data=list(sorter(data,lambda d:d[target],reverse))
        # print([d["Email"] for d in sorted_data])
        return sorted_data




import src.utility.logs
__version__="0.0.1"
from pathlib import Path
from src.app.cli import main
base_path=Path.cwd()
test_project=base_path.parent.joinpath("pandas")
test_project2=base_path.parent.joinpath("QualiTag")
test_project3=base_path.parent.joinpath("sudoku-gen-n-solve")
test_project4=base_path.parent.joinpath("tensorflow")
if __name__ =="__main__":
    main([base_path.as_posix()],env="PROD")



import dash
from dash import dcc,callback,Input,Output,no_update,set_props,State,clientside_callback,Patch,ctx,MATCH
from dash.exceptions import PreventUpdate
import dash.html as html
import plotly.express as px
import pandas as pd
from src.app.helper import create_info_card_columns
from repository_miner import RepoMiner
from src._internal.data_typing import Author
import dash_bootstrap_components as dbc
from dash_ag_grid import AgGrid
from src.gui import AuthorDisplayerAIO,CommitDisplayerAIO
from logging import getLogger
logger=getLogger("mainpage")
dash.register_page(__name__,"/")
common_labels={"date":"Date","commit_count":"Number of commits","author_email":"Author's email","author_name":"Author's name","dow":"Day of the week"}
truck_facto_modal=dbc.Modal(
        [
                dbc.ModalHeader("How we calculate the truck factor"),
                dbc.ModalBody("The truck factor is calculated through a naive version of the AVL algorithm for truck factor calculation; the DOA (Degree of Authorship) used for truck factor calculation is obtained evaluating the number of non-whitespace commits authored by each author (it will not take into account the number of lines changed) for each file of the project. The final number it is the result of an operation of thresholding for which we discard all DOA normalized values inferior to 0.75, the resulting DOAs obtained from the filtering process are then used to estabilish the number of file authored by each author in order to lazily remove each author from the calculation until at least 50% of project's file are 'orphans'(no file author alive). The number of author to remove in order to satisfy the previous condition is the effective truck factor calculated for the project" ),
        ],"truck_factor_modal",is_open=False
)

column_defs_commits=[
                {"field": "commit_hash", 'headerName': 'Commit Hash',"filter": "agTextColumnFilter"},
                {"field": "author_name",'headerName': 'Author Name',"filter": "agTextColumnFilter"},
                {
                        "field":"date",
                        "headerName":"Date",
                        "headerName": "Date",
                        "filter": "agDateColumnFilter",
                        "sortable":True,
                }
        ]

column_defs_authors=[
        {"field": "email", 'headerName': 'Author Email',"filter": "agTextColumnFilter"},
        {"field": "name",'headerName': 'Author Name',"filter": "agTextColumnFilter"},
        {"field": "commits_authored",'headerName': 'Commits Authored',"sortable":True},
        {"field": "files_authored",'headerName': 'Files Authored',"sortable":True},
]

layout = dbc.Container([
        truck_facto_modal,
        dbc.Tooltip("Click on the commit hash for commit description",target="commit_tooltip",trigger="legacy",is_open=False,id="commit_tooltip_info"),
        dbc.Tooltip("Choose a range to filter out all the authors that are out of commit count range",target="commit_slider",is_open=False,placement="top-end",trigger="hover legacy"),

         dbc.Modal([
                dbc.ModalHeader([html.I(className="bi bi-git h3 pe-3"),html.Span(f"Commit: ",className="fw-bold"),html.Span(id="commit_modal_header",className="fw-bold")]),
                dbc.ModalBody([
                    dbc.Container([
                        html.P([html.Span("Commit Message: ",className="fw-bold"), html.Span(id="commit_modal_message")] ),
                        html.P([html.Span("Commit Author: ",className="fw-bold"), html.Span(id="commit_modal_author")]),
                        html.P([html.Span("Complete hash string: ",className="fw-bold"),html.Span(id="commit_modal_hash")]),
                        html.P([html.Span("Created at: ",className="fw-bold") ,html.Span(id="commit_modal_date")]),
                        
                    ]),
                ])
            ],id="commit_modal",size="lg"),
        dbc.Row(id="repo_general_overview_row",class_name="mb-3"),
        dbc.Row(id="repo_graph_row",children=[
                
                dbc.Col(
                        [
                                dcc.Loading([
                                        dbc.Card([
                                                dbc.CardHeader(children=[
                                                                        html.I(className="bi bi-clipboard2-data pe-3 d-inline h2"),html.Span("Repository analysis",className="fw-bold h2"),
                                                                ]),
                                                dbc.CardBody(id="repo_info",children=[
                                                        dcc.Loading(id="author_overview_loader",children=[
                                                         dcc.Graph(id="repo_info_graph")
                                                                ],
                                                                overlay_style={"visibility":"visible", "filter": "blur(2px)"},
                                                                ),
                                                                
                                                ]),
                                        ])
                                ])
                        ]
                ,width=6,align="start"
                ),
                dbc.Col(
                        [
                                dbc.Card([
                                                dbc.CardHeader(id="contribution_info_header",children=[
                                                        html.I(className="bi bi-truck d-inline h2 pe-3"),
                                                        html.H4(f"Truck factor",className="d-inline fw-bold h2")
                                                ]),
                                                dbc.CardBody(id="contribution_info",children=[
                                                        dcc.Loading(id="author_overview_loader",children=[
                                                         dcc.Graph(id="contribution_graph")
                                                        ],
                                                        overlay_style={"visibility":"visible", "filter": "blur(2px)"},
                                                        ),
                                                        
                                                ]),
                                        ])                                
                        ]
                ,width=6,align="start"
                ),
                ],class_name="pb-4"),
        dbc.Tabs([
                dbc.Tab(
                        [
                        dbc.Row(id="author_graph_row",children=[           
                        dbc.Col(
                                [
                                        dbc.RadioItems(id="x_picker",options=[{"label":"Day of week","value":"dow"},{"label":"Per date","value":"date"}],value="dow",inline=True, switch=True),
                                        dcc.Loading(id="author_loader_graph",
                                        children=[dcc.Graph(id="graph",className="h-100")],
                                        overlay_style={"visibility":"visible", "filter": "blur(2px)"},
                                ),
                                ],width=7,align="center"),
                        dbc.Col([
                                 dbc.RadioItems(
                                        options=[
                                                {"label": "Pie graph", "value": "pie"},
                                                {"label": "Scatter Plot", "value": "scatter"},
                                        ],
                                        value="scatter",
                                        id="author_graph_picker",
                                        switch=True,
                                        inline=True,
                                        ),
                                dcc.Loading(id="author_overview_loader",children=[
                                                dcc.Graph(id="author_overview")
                                        ],
                                        overlay_style={"visibility":"visible", "filter": "blur(2px)"},
                                        ),
                                
                                ],width=4),
                        dbc.Col(width=1,children=[
                                dcc.RangeSlider(min=1,max=10,id="commit_slider",value=[1,10],vertical=True,tooltip={"placement": "bottom", "always_visible": False})
                        ]),
                                
                        ],justify="center"),
                dbc.Row([
                        
                ])
                ],label="History overview"
                ),
                dbc.Tab(label="Authors",children=[
                                        dbc.Row(children=[
                                                dbc.Col(width=12,align="center",id="authors_tab",children=[
                                                AgGrid(
                                                id="authors_table",
                                                columnDefs=column_defs_authors,
                                                columnSize="responsiveSizeToFit",
                                                defaultColDef={"sortable":False,"resizable":True},
                                                dashGridOptions={"pagination": True, "animateRows": False},
                                                )
                                                        ]),
                                        ],justify="center"),]),
                dbc.Tab(label="Commits",children=[
                                        html.I(id="commit_tooltip",className="bi bi-question fw-bold fs-3 clickable"),                                                  
                                        dbc.Row(children=[
                                                dbc.Col(width=12,align="center",id="commits_tab",children=[
                                                AgGrid(
                                                id="commits_table",
                                                columnDefs=column_defs_commits,
                                                columnSize="responsiveSizeToFit",
                                                defaultColDef={"sortable":False,"resizable":True},
                                                dashGridOptions={"pagination": True, "animateRows": False},
                                                )
                                                        ]),
                                        ],justify="center"),
                                        ]),
                ]),
        # html.Div(id="test-div")
],fluid=True,className="p-10")

@callback(
        Output("repo_general_overview_row","children"),
        Input("authors_cache","data"),
        State("contribution_cache","data"),
        State("branch_picker","value"),
        State("repo_path","data"),
)
def populate_generale_info(authors,contributions,branch,path,):
        rp=RepoMiner(path)
        contrs=pd.DataFrame(contributions)
        num_commits=rp.n_commits()
        current_head=rp.git.rev_parse(["--abbrev-ref","HEAD"]) if not branch else branch
        num_authors=len(authors)
        current_commit=rp.get_commit(branch if branch else current_head)
        local_branches_num=len(rp.local_branches_list())
        tag_num=len(list(rp.get_tags()))
        texts={
                "Total number of commits":f"{num_commits}",
                "Total number of authors":f"{num_authors}",
                # f"Current head of repository: {current_head}",
                "Number of branches":f"{local_branches_num}",
                "Number of tags":f"{tag_num}",
                "Number of files of interest":f"{str(len(contrs['fname'].unique()))}",
                "Last reachable commit":CommitDisplayerAIO(current_commit).create_comp()
        }
        icons={
                "Total number of commits":html.I(className="bi bi-graph-up pe-3 d-inline ms-2 h4 fw-bold text-start"),
                "Total number of authors":html.I(className="bi bi-pen-fill d-inline ms-2 pe-3"),
                "Number of branches":html.I(className="bi bi-bezier2 pe-3 d-inline ms-2"),
                "Number of tags":html.I(className="bi bi bi-tags pe-3 d-inline ms-2"),
                "Number of files of interest":html.I(className="bi bi-search pe-3 d-inline ms-2"),
                "Last reachable commit":html.I(className="bi bi-code-slash pe-3 d-inline ms-2")
        }
        cards:list=create_info_card_columns(texts,icons)
        return cards
        
       
        

@callback(
        Output("repo_info_graph","figure"),
        Input("contribution_cache","data"),
)
def populate_repo_info(contributions):
        def extract_suffix(filename:str)->str:
                try:
                        return filename.rsplit(".",maxsplit=1)[1]
                except IndexError:
                        return "no extension"
        contrs=pd.DataFrame(contributions)
        fnames=contrs["fname"].unique()
        suffixes=[extract_suffix(f) for f in fnames]
        df=pd.DataFrame(dict(fname=fnames,suffix=suffixes))
        df=df.groupby("suffix",as_index=False).count()
        # print(df.head(999))
        tot:int=df["fname"].sum()
        th_percentage=3*tot/100
        df.loc[df['fname'] < th_percentage, 'suffix'] = 'Other'
        fig = px.pie(df, values='fname', names='suffix',labels={"fname":"files"},hole=0.5)
        fig.update_layout(annotations=[dict(text='Languages',
                      font_size=20, showarrow=False, xanchor="center")])
        return fig

@callback(
        Output("commits_table","rowData"),
        Input("branch_cache","data")
)
def populate_commits_tab(data):
        if not data:
                return no_update
        df=pd.DataFrame(data)
        df["date"]=pd.to_datetime(df["date"])
        return df.to_dict("records")

@callback(
        Output("commit_modal_header","children"),
        Output("commit_modal_message","children"),
        Output("commit_modal_author","children"),
        Output("commit_modal_hash","children"),
        Output("commit_modal_date","children"),
        Output("commit_modal","is_open"),
        Input("commits_table","cellClicked"),
        State("branch_cache","data"),
        prevent_initial_call=True
)
def listen_commits_tab_click(cell,data):
        if cell["colId"]!="commit_hash":
                raise PreventUpdate()
        df=pd.DataFrame(data)
        hash=cell["value"]
        commit:pd.Series=df.loc[df["commit_hash"]==hash].iloc[0]
        return hash[:7],commit["subject"],f"{commit['author_name']} <{commit['author_email']}>",hash,commit["date"],True
        
@callback(
        Output("authors_table","rowData"),
        Input("contribution_cache","data"),
        Input("authors_cache","data")
)
def populate_authors_tab(contributions,data,doa_th=0.75):
        if not contributions:
                return no_update
        contr=pd.DataFrame(contributions)
        authors=pd.DataFrame(data)
        contr=contr[contr.DOA >= doa_th]
        contr=contr.groupby("author",as_index=False).count()
        contr.rename(columns={"author":"name"},inplace=True)
        authors=authors.join(contr.set_index("name"),rsuffix="contr",on="name",validate="m:1")
        authors.rename(columns={"DOA":"files_authored"},inplace=True)
        authors["commits_authored"]=authors["commits_authored"].map(lambda a: len(a))
        authors.fillna(0,inplace=True)
        return authors.to_dict("records")

@callback(
        Output("graph","figure"),
        Input("x_picker","value"),
        Input("branch_cache","data"),
        State("branch_picker","value"),
)
def update_count_graph(pick,data,branch):
        if not data:
                return no_update
        commit_df=pd.DataFrame(data)
        if pick =="dow":
                count_df=commit_df.groupby(["dow","dow_n"])
                count_df=count_df.size().reset_index(name="commit_count")
                count_df.sort_values("dow_n",ascending=True,inplace=True)
                fig=px.bar(count_df,x=pick,y="commit_count",labels=common_labels,title=f"Commit Distribution {branch if branch else ''}")
        else:
                count_df=commit_df.groupby(["date"]).size().reset_index(name="commit_count")
                fig=px.area(count_df,hover_data=["date"],x=pick,y="commit_count",labels=common_labels,title=f"Commit Distribution {branch if branch else ''}")
        return fig

@callback(
        Output("author_overview","figure"),
        Input("authors_cache","data"),
        Input("branch_cache","data"),
        Input("author_graph_picker","value"),
        Input("commit_slider","value"),
        State("contribution_cache","data"),
)
def update_pie_graph(data,b_cache,pick,range,contribution):
        df=pd.DataFrame(data)
        b_df=pd.DataFrame(b_cache)
        allowed_commits=set(b_df["commit_hash"].to_list())
        df["commits_authored"]=df["commits_authored"].apply(lambda r: set(r).intersection(allowed_commits))
        df["contributions"]=df["commits_authored"].apply(lambda r: len(r))
        df=df.groupby("name",as_index=False).sum(True)
        min,max=range
        mask = (df['contributions'] >= min) & (df['contributions'] <= max)
        df=df[mask]
        if pick =="pie":
                tot:int=df["contributions"].sum()
                th_percentage=2*tot/100
                df.loc[df['contributions'] < th_percentage, 'name'] = 'Minor contributors total effort'
                fig = px.pie(df, values='contributions', names='name', title='Authors commit distribution')
        else:
                contrs=pd.DataFrame(contribution)
                contrs=contrs.groupby("author").sum()
                df=df.join(contrs,rsuffix="contr",on="name",validate="m:1").reset_index()
                fig = px.scatter(df,x="name",y="contributions",color="tot_contributions",title='Authors commit distribution', labels={"contributions":"commit count","tot_contributions":"contributions"})
                fig.update_xaxes(showticklabels=False)
                fig.update_yaxes(gridcolor='lightgrey')
                fig.update_layout(
                        plot_bgcolor='white'
                )
        return fig
@callback(
      Output("commit_slider","max"),
      Output("commit_slider","value"),  
      Input("authors_cache","data"),
      Input("branch_cache","data"),         
)

def setup_slider(data,b_cache):
        df=pd.DataFrame(data,)
        b_df=pd.DataFrame(b_cache)
        allowed_commits=set(b_df["commit_hash"].to_list())
        df["commits_authored"]=df["commits_authored"].apply(lambda r: set(r).intersection(allowed_commits))
        df["contributions"]=df["commits_authored"].apply(lambda r: len(r))
        df=df.groupby("name",as_index=False).sum(True)
        max= df["contributions"].max()
        return max , [1,max]

@callback(
        Output("contribution_graph","figure"),
        Input("truck_cache","data"),
        State("contribution_cache","data"),
        prevent_inital_call=True
)
def populate_contributors(tf,contributions,th=0.75):
        if not contributions:
                return no_update
        contrs=pd.DataFrame(contributions)
        contrs=contrs.loc[contrs["DOA"]>=th]
        top=contrs.groupby("author",as_index=False).count()
        top=top.sort_values("DOA",ascending=False).head(tf)
        tot:int=top["DOA"].sum()
        th_percentage=3*tot/100
        top.loc[top['DOA'] < th_percentage, 'author'] = 'Other contributors'
        # print(top.head(tf))
        fig = px.pie(top, values='DOA', names='author',labels={"DOA":"files authored"},hole=0.4)
        fig.update_layout(annotations=[dict(text=tf,
                      font_size=20, showarrow=False, xanchor="center")])
        return fig

@callback(
        Output("commit_tooltip_info","is_open"),
        Input("commit_tooltip","n_clicks"),
        prevent_inital_call=True
)
def open_toast_commit(_):
        if _==None:
                return no_update
        return True


from pytest import mark,fixture,raises
from dataclasses import dataclass,field
from src.utility.helper import *
from src.utility.filters import filter,Filter,FilterFactory,sorter_provider,sorter
from src.app.helper import *
from src._internal.file_parser import DEFAULT_SATD_HIGHLIHGTER
from repository_miner import RepoMiner
from pathlib import Path
main_path=Path.cwd()
test_project=main_path.parent.joinpath("pandas")
@fixture
def git():
    return RepoMiner(main_path.as_posix())

@dataclass
class SerializableTestDataClass():
    a:str
    b:int
    d:dict
    c:list = field(default_factory=lambda: [])
    
    
class SerializableTestClass():
    
    def __init__(self,
                a:str
                ,b:int
                ,c:list
                ,d:dict):
        self.a=a
        self.b=b
        self.c=c
        self.d=d
    

@mark.parametrize("files",[("src/app.py","tests/report.txt","tests/test_dummy.py","tests/test_file_parser.py"),(".github/workflows/test-dev.yml",".github/workflows/testpypi_publish.yml",".gitignore","LICENSE","RAD.docx","README.md","main.py","pyproject.toml","requirements.txt","src/_internal/__init__.py","src/_internal/data_preprocessing.py","src/_internal/data_typing.py","src/_internal/file_parser.py","src/_internal/git_mining.py","src/_internal/info/ext.json","src/app/__init__.py","src/app/app.py","src/app/cli.py","src/gui/__init__.py","src/gui/components.py","src/gui/pages/homepage.py","src/utility/__init__.py","src/utility/logging_configs/logs_config_file.json","src/utility/logging_configs/logs_config_file_old.json","src/utility/logs.py","tests/test_cli.py","tests/test_data_preprocessing.py","tests/test_dummy.py","tests/test_file_parser.py","tests/test_git_miner.py")])
def test_infer_programming_language(files):
    assert infer_programming_language(files)==[".py"]

@mark.parametrize("numbers,comparison,expected",[
    ([1,5,2,6,3,8,34,6,3,7,3,789,3,5,3,55],33,3),
    ([1,5,2,6,3,8,34,6,3,7,3,789,3,5,3],33,2),
    ([1,5,2,6,3,8,6,3,7,3,3,5,3],33,0),
])
def test_filter_functions(numbers,comparison,expected):
    @filter("test_func")
    def test(value,comparison):
        return value > comparison
    fil:Filter = FilterFactory.create_filter("test_func")
    filtered=list(fil.run(numbers,comparison))
    assert len(filtered)==expected
    
@mark.parametrize("commit_sha,path_filter,expected",[
    ("HEAD",{"tests/test_cli.py"},True),
    ("",{"tests/test_cli.py"},False),
    (None,{"tests/test_cli.py"},False),
    ("HEAD",{},True),
])
def test_build_tree(git,commit_sha,path_filter,expected):
    if expected:
        build_tree_structure(git,commit_sha,path_filter)
    else:
        with raises(Exception):
            build_tree_structure(git,commit_sha,path_filter)

def test_retrieve_satds(git):
    files=[]
    for hg in DEFAULT_SATD_HIGHLIHGTER:
        files.extend(git.git.grep(["-l","-E",f'\"[^a-zA-Z0-9_]+\\s*({hg}).+\"']).split("\n"))
    satds=retrieve_SATDs(git,DEFAULT_SATD_HIGHLIHGTER)
    assert set(files) == set(satds.keys())

def test_retreive_contribution():
    df,tf=retrieve_contribution_data(main_path)
    print(tf)

def test_parallel_commit_retrieval(git):
    cl=git.retrieve_commit_list(merges=True)
    assert sorted(cl,key=lambda c: c.abbr_hash) ==sorted(parallel_commit_retrievial(git),key=lambda c: c.abbr_hash)
    
def test_dataframe_serialization():
    o1=SerializableTestDataClass("ciao",1,{"b":["a"]},["giovanni"])
    o2=SerializableTestClass("abc",2,["tommaso"],{})
    d1=get_dataframe(o1)
    d2=get_dataframe(o2)
    data1=dict(
        a=["ciao"],
        b=[1],
        d=[{"b":["a"]}],
        c=[["giovanni"]]
    )
    data2=dict(
        a=["abc"],
        b=[2],
        c=[["tommaso"]],
        d=[{}]
    )
    df1=pd.DataFrame(data1)
    df2=pd.DataFrame(data2)
    pd.testing.assert_frame_equal(df1,d1)
    pd.testing.assert_frame_equal(df2,d2)



import pandas as pd
from typing import Iterable
# from inspect import get_annotations
from src._internal.data_typing import Author
from time import strftime,gmtime
from datetime import date
from functools import cache
from logging import getLogger
import time
from datetime import datetime
logger=getLogger("Data Preprocessing")

def unixTimeMillis(dt:pd.Timestamp):
    ''' Convert datetime to unix timestamp '''
    return int(dt.timestamp())

def unixToDatetime(unix:int)->pd.Timestamp:
    ''' Convert unix timestamp to datetime. '''
    return pd.to_datetime(unix,unit='s',utc=True)
def getMaxMinMarks(start_date:pd.Timestamp,end_date:pd.Timestamp):
    ''' Returns the marks for labeling. 
    '''
    date_format='%Y-%m-%d'
    result={}
    result[unixTimeMillis(start_date)] = str(start_date.strftime(date_format))
    result[unixTimeMillis(end_date)] = str(end_date.strftime(date_format))
    return result
def getMarks(dates:Iterable[pd.Timestamp], Nth=100):
    ''' Returns the marks for labeling. 
        Every Nth value will be used.
    '''

    result = {}
    for i, date in enumerate(dates):
        if(i%Nth == 1):
            # Append value to dict
            result[unixTimeMillis(date)] = str(date.strftime('%Y-%m-%d'))
    return result


from pytest import mark,raises,fixture
from src._internal.data_typing import Author,File,Folder,TreeStructure
from src._internal.data_preprocessing import getMaxMinMarks,unixTimeMillis,unixToDatetime,getMarks
from datetime import date
from src._internal.exceptions import ObjectNotInTreeError,PathNotAvailableError
import datetime as dt
import pandas as pd
from utility import logs as log
from pathlib import Path
import logging
log.setup_logging()
logger=logging.getLogger("Data preprocessor tester")

@fixture
def tree():
    fo=Folder("folder1",dict(file1=File("file1",10,"asdasdasd"),folder2=Folder("folder2",dict(),"kljfjf")),"kljklfjlfkj")
    fi=File("file2",size=12,hash_string="dsfdgds")
    return TreeStructure("sajdhasjdhk",[fo,fi])

today=int(dt.datetime.today().timestamp())-1000
# pd.to_datetime(,unit='s')
later=(dt.datetime.now().timestamp())
def test_marks():
    assert len(getMarks([pd.to_datetime(today,unit='s'),pd.to_datetime(later,unit='s')],2).keys())==1
    assert len(getMarks([pd.to_datetime(today,unit='s'),pd.to_datetime(later,unit='s')],1).keys())==0
    with raises(ZeroDivisionError):
        len(getMarks([pd.to_datetime(today,unit='s'),pd.to_datetime(later,unit='s')],0).keys())==2
def test_max_min_marks():
    assert len(getMaxMinMarks(pd.to_datetime(today,unit='s'),pd.to_datetime(later,unit='s')).keys())==2

def test_unixtimemillis():
    assert int(unixTimeMillis(pd.to_datetime(today,unit='s'))/10000)==int(today/10000)
    logger.debug(unixTimeMillis(pd.to_datetime(-1,unit='s')))
    
def test_unixToDatetime():
    assert unixToDatetime(unixTimeMillis(pd.to_datetime(today,unit='s')))==pd.to_datetime(today,unit='s',utc=True)
    logger.debug(unixToDatetime(-1))
    
tree_walk_args=[
    ("file2",False,False,[File("file2",size=12,hash_string="dsfdgds")]),
    ("folder1",False,False,[File("file1",10,"asdasdasd"),Folder("folder2",dict(),"kljfjf")]),
    ("folder1",False,True,[Folder("folder2",dict(),"kljfjf")]),
    ("folder1",True,False,[File("file1",10,"asdasdasd")]),
    ("file1",False,False,None),
    ("file1",True,True,None),
    ("file3",False,False,None),
]
@mark.parametrize("name,fo,do,expected",tree_walk_args)
def test_walk_folder(tree,name,fo,do,expected):
    if expected:
        assert set(tree.walk_folder(name,fo,do))==set(expected)
    else:
        with raises((ObjectNotInTreeError,ValueError,TypeError)):
            list(tree.walk_folder(name,fo,do))

tree_walk=[
    (False,False,[File("file1",10,"asdasdasd"),File("file2",size=12,hash_string="dsfdgds"),Folder("folder1",dict(file1=File("file1",10,"asdasdasd"),folder2=Folder("folder2",dict(),"kljfjf")),"kljklfjlfkj"),Folder("folder2",dict(),"kljfjf")]),
    (False,True,[Folder("folder2",dict(),"kljfjf"),Folder("folder1",dict(file1=File("file1",10,"asdasdasd"),folder2=Folder("folder2",dict(),"kljfjf")),"kljklfjlfkj")]),
    (True,False,[File("file2",size=12,hash_string="dsfdgds"),File("file1",10,"asdasdasd")]),
    (True,True,None),
]
@mark.parametrize("fo,do,expected",tree_walk)
def test_walk(tree,fo,do,expected):
    if expected:
        assert set(list([o for p,o in tree.walk(fo,do)]))==set(expected)
    else:
        with raises((ObjectNotInTreeError,ValueError,TypeError)):
            list([o for p,o in tree.walk(fo,do)])
tree_find_args=[
    ("file2","folder",[]),
    ("folder1","folder",[Folder("folder1",dict(file1=File("file1",10,"asdasdasd"),folder2=Folder("folder2",dict(),"kljfjf")),"kljklfjlfkj")]),
    ("file1","file",[File("file1",10,"asdasdasd")]),
    ("file2","file",[File("file2",size=12,hash_string="dsfdgds")]),
    ("file1",None,[File("file1",10,"asdasdasd")]),
    ("file1","d",None),
    ("file1","folder",[]),
    
]
@mark.parametrize("name,tp,expected",tree_find_args)
def test_find(tree,name,tp,expected):
    if expected!=None:
        assert set(tree.find(name,tp))==set(expected)
    else:
        with raises(TypeError):
            list(tree.find(name,tp))

def test_get_path(tree):
    for p,o in tree.walk():
        if o.name=="folder2":
            assert p=="folder1"
            
@mark.parametrize("path,expected",[("folder1/folder2",Folder("folder2",dict(),"kljfjf")),("file2",File("file2",size=12,hash_string="dsfdgds")),("folder1/file2",None),("folder2",None)])
def test_get(tree,path,expected):
    if expected:
        tree.get(path)==expected
    else:
        with raises(ObjectNotInTreeError):
            tree.get(path)

build_args=[
    ("a/b/c",File("c",0,"asd"),True,File("c",0,"asd")),
    ("a/b/c",File("c",0,"asd"),False,None),
    ("a/b/c",Folder("c",dict(),"asd"),True,Folder("c",dict(),"asd")),
    ("a/b/c",Folder("c",dict(),"asd"),False,None),
    ("c",Folder("c",dict(),"asd"),False,Folder("c",dict(),"asd")),
    ("c",Folder("c",dict(c=File("c",0,"asd")),"asd"),False,Folder("c",dict(c=File("c",0,"asd")),"asd")),
    ("",Folder("c",dict(),"asd"),False,None),
    ("a/b/c",Author("c","d",["asd"]),False,None),
    ("a/b/c",None,False,None),
]
@mark.parametrize("path,obj,mkdir,expected",build_args)
def test_build_tree(path,obj,mkdir,expected):
    if expected:
        tree=TreeStructure(hash="asjd")
        tree.build(path,obj,mkdir)
        assert next(tree.find(expected.name))==expected
    else:
        with raises((ValueError,ObjectNotInTreeError,TypeError)):
            tree=TreeStructure(hash="asjd")
            tree.build(path,obj,mkdir)
            
static_build_args=[
    ("hash","a/b/c",File("c",0,"asd"),True,File("c",0,"asd")),
    ("hash","a/b/c",File("c",0,"asd"),False,None),
    ("hash","a/b/c",Folder("c",dict(),"asd"),True,Folder("c",dict(),"asd")),
    ("hash","a/b/c",Folder("c",dict(),"asd"),False,None),
    ("hash","c",Folder("c",dict(),"asd"),False,Folder("c",dict(),"asd")),
    ("hash","c",Folder("c",dict(c=File("c",0,"asd")),"asd"),False,Folder("c",dict(c=File("c",0,"asd")),"asd")),
    ("hash","",Folder("c",dict(),"asd"),False,None),
    ("hash","a/b/c",Author("c","d",["asd"]),False,None),
    ("hash","a/b/c",None,False,None),
]
@mark.parametrize("hash,path,obj,mkdir,expected",static_build_args)
def test_static_build_tree(hash,path,obj,mkdir,expected):
    if expected:
        tree=TreeStructure.build_tree(hash,path,obj,mkdir)
        assert next(tree.find(expected.name))==expected
    else:
        with raises((ValueError,ObjectNotInTreeError,TypeError)):
            tree=TreeStructure.build_tree(hash,path,obj,mkdir)
            
def test_insertion_build():
    tree =TreeStructure(hash="asjd",content=[Folder("c",dict(d=File("d",0,"aswd"),a=Folder("a",dict(),"sdhf")),"asd")])
    with raises(PathNotAvailableError):
        tree.build("c/d",File("d",0,"sdf"))
    with raises(PathNotAvailableError): 
        tree.build("c/a",Folder("a",0,"sdf"))



from pathlib import Path
from typing import Union,Generator,Iterator,Literal
import re
import os
import logging
from .data_typing import ACCEPTED_EXTENSIONS
logger = logging.getLogger("File Parser")
DEFAULT_SATD_HIGHLIHGTER={"TODO","FIXME","HACK","XXX"}
def fetch_source_files(project_path:Union[Path|str],extensions:set[str],exclude_dirs:set[str]=[".venv",".git",".pytest_cache"])->Generator[Path,None,None]:
    # info("Entered fetch_source_files function")
    path = project_path
    if isinstance(path,str):
        path = Path(project_path)
    if not path.exists():
        logger.critical("Path does not exist")
        raise FileNotFoundError("Path does not exist")
    # info(f"Path {path.as_posix()} Exists")
    if not path.is_dir():
        logger.critical("Path is not a directory")
        raise NotADirectoryError("Path is not a directory")
    # info(f"Path {path.as_posix()} is a directory")
    for item in path.iterdir():
        if item.is_dir() and item.name in exclude_dirs:
            continue
        elif item.is_dir():
            for i in fetch_source_files(item,extensions):
                yield i
        elif item.suffix in extensions:
            yield item
# TODO could be optimized using multiprocessing
def _comment_finder(text:Union[str,list[str]],single_line_pattern:list[str],multi_line_pattern:list[str])->list[tuple[int,int,str]]:
    content=[]
    txt=""
    multi_line:Iterator[re.Match[str]] = ()
    comments:list[tuple[int,int,str]] = []
    if isinstance(text,str):
        txt=text
        content=text.splitlines()
    else:
        txt="".join(text)
        content=text
    for i, line in enumerate(content, 1):
        for pattern in single_line_pattern:
            single_line = re.findall(pattern, line)
            if single_line:
                comments.append((i,i, single_line[0]))
    for pattern in multi_line_pattern:
        multi_line = re.finditer(pattern, txt)
        for match in multi_line:
            matched_string=match.group()
            start_line = txt[:match.start()].count('\n') + 1
            end_line = matched_string.count('\n') + start_line
            comments.append((start_line,end_line, matched_string))
    return comments

def find_file_comments_with_locations(filename:Union[str,Path])->list[tuple[int,int,str]]:
    """Find all comments inside a file discriminating comments markers from code language inferred from the file extension 

    Args:
        filename (Union[str,Path]): path to the file

    Raises:
        FileNotFoundError: if file is not readable

    Returns:
        list[tuple[int,int,str]]: a list of triplets organized as follows: comment start n° line, comment end n° line, comment string
    """    
    filepath=filename
    if isinstance(filename,Path):
        filepath=filename.as_posix()
    if not os.path.isfile(filepath):
        logger.critical(f"Path {filepath} is not a file or does not exist")
        raise FileNotFoundError(f"Path {filepath} is not a file or does not exist")
    _, ext = os.path.splitext(filepath)
    
    with open(filepath, 'r',encoding="utf-8") as file:
        content = file.readlines()
        
    return find_comments_with_locations(content,ext=ext)

def find_comments_with_locations(text:Union[str|list[str]],ext:str)->list[tuple[int,int,str]]:
    """Find all comments inside a string discriminating comments markers from code language inferred from the file extension 

    Args:
        text (Union[str | list[str]]): text to parse
        ext (ACCEPTED_EXTENSIONS): virtual extension of the text 

    Returns:
        list[tuple[int,int,str]]: a list of triplets organized as follows: comment start n° line, comment end n° line, comment string
    """    
    content=text
    if isinstance(text,str):
        content=re.split(string=text,pattern=r'\r\n|\n|\r')
    comments:list[tuple[int,int,str]] = []
    
    if ext in ['.py', '.rb',".sh"]:
        if ext ==".sh":
            comments =_comment_finder(content,[r'#.*'],[r':\'[\s\S]*?\''])
        else:
            comments =_comment_finder(content,[r'#.*'],[r'"""[\s\S]*?"""'])
            
    elif ext in ['.js', '.java', '.c', '.cpp',".php",".sql"]:
        if ext == ".php":
            comments=_comment_finder(content,[r'//.*',r'#.*'],[r'/\*[\s\S]*?\*/'])
        elif ext == ".sql":
            comments=_comment_finder(content,[r'//.*',r'--.*'],[r'/\*[\s\S]*?\*/'])
        else:
            comments=_comment_finder(content,[r'//.*'],[r'/\*[\s\S]*?\*/'])
        
    elif ext == '.html':
        # HTML comments
        comments=_comment_finder(content,[],[r'<!--[\s\S]*?-->'])
    else:
        logger.debug(f"Path {ext} not expected for comment finding",extra={"extension":ext})
    # Add more language-specific rules as needed
    return comments

def _find_satd_inline(text:str,tags):
    for tag in tags:
        mt:re.Match = re.match(f"^\\W+\\s*({tag}.*)",text)
        if mt and isinstance(mt,re.Match):
            return mt.group(1)
def _find_satd(comments:list[tuple[int,int,str]],tags:list[str])->dict[int,str]:
    satds:dict[int,str]=dict()
    for start,end,comment in comments:
        if start == end:
            txt=_find_satd_inline(comment,tags)
            if txt:
                satds[start]=txt
        else:
            content=re.split(string=comment,pattern=r'\r\n|\n|\r')
            for i,cont in enumerate(content,1):
                txt=_find_satd_inline(cont,tags)
                if txt:
                    satds[start+i]=txt
    return satds
def find_satd_file(filepath:Union[Path|str],tags:set[str]=DEFAULT_SATD_HIGHLIHGTER)->dict[int,str]:
    """Finds all SATD in a file

    Args:
        filepath (Union[Path | str]): path to the file
        tags (set[str], optional): tags used to mark SATD. Defaults to {"TODO","FIXME","HACK","XXX"}.

    Returns:
        dict[int,str]: dictionary with line as key and SATD content as value
    """    
    comments=find_file_comments_with_locations(filename=filepath)
    logger.debug(f"Found comments in file {filepath.as_posix() if isinstance(filepath,Path) else filepath}",extra={"comments":comments})
    return _find_satd(comments,tags)
def find_satd(text:str,extension:str,tags:set[str]=DEFAULT_SATD_HIGHLIHGTER)->dict[int,str]:
    """Finds all SATD in a text

    Args:
        text str: text to analyze.
        extension (ACCEPTED_EXTENSIONS): extention used to infer the programming language.
        tags (set[str], optional): tags used to mark SATD. Defaults to {"TODO","FIXME","HACK","XXX"}.

    Returns:
        dict[int,str]: dictionary with line as key and SATD content as value
    """    
    comments=find_comments_with_locations(ext=extension,text=text)
    logger.debug("Found comments in text",extra={"comments":comments})
    return _find_satd(comments,tags)


from argparse import ArgumentParser,FileType
from .app import start_app
from repository_miner.git import Git
from repository_miner.exceptions import GitNotFoundException,NotGitRepositoryError
from pathlib import Path
from typing import Optional,Sequence
from sys import version_info
import subprocess
import logging
import os
from src.utility._version import __version__
from src.utility.logs import setup_logging
logger=logging.getLogger("cli")
def main(args:Optional[Sequence[str]]=None,cicd_test:bool=False,env:str="PROD"):
    setup_logging(env=env)
    if version_info.minor<10:
        logger.exception(f"System Python version {version_info.major}.{version_info.minor} < Python3.10.x . It is mandatory to at least use Python versions > 3.10.x for a correct usage of the application")
        exit(5)
    parser=ArgumentParser(prog="project-viewer")
    parser.add_argument('dir',nargs='?', default=Path.cwd().as_posix(), type=str)
    parser.add_argument('-v',"--version",action='store_true')
    parsed_args=parser.parse_args(args)
    if parsed_args.version==True:
        print("project-viewer "+__version__)
        exit(0)
    namespace=parsed_args.dir
    dir=str(namespace)
    if dir==".":
        dir=Path.cwd().as_posix()
    logger.info(f"Opening git repository at {dir}")
    if not Path(dir).is_dir():
        logger.error(f"Path {dir} is not a directory")
        exit(1)
    #check if git is installed
    try:
        Git(Path(dir).resolve().absolute().as_posix())
    except (GitNotFoundException,NotGitRepositoryError) as p:
        logger.error(p)
        exit(3)
    # git_dir=Path(dir).joinpath(".git")
    # #check if dir is a git repository
    # if not git_dir.is_dir():
    #     logger.error(f"Chosen directory is not a git repository")
    #     exit(2)
    # try:
    #     subprocess.check_call(["git","-C",dir,"rev-parse","--git-dir"],stderr=open(os.devnull, 'wb'),stdout=open(os.devnull, 'wb'))
    # except subprocess.CalledProcessError:
    #     logger.error("Git repo is corrupted, check for your git config files")
    #     exit(3)
    logger.info(f"Starting application")
    start_app(Path(dir).resolve().absolute().as_posix(),cicd_test,env=env)
    # find_setd()


from logging import getLogger
from typing import Iterable,Literal,Union,Callable,Type
from math import log1p
import json
import pandas as pd
import unicodedata
logger=getLogger("helper")

def infer_programming_language(files:Iterable[str],threshold:float=0.35)->list[str]:
        suffix_count:dict[str,int]=dict()
        fs=set(files)
        tot_files=len(fs)
        ret_suffixes=list()
        for file in fs:
            try:
                suffix=file.rsplit(".",maxsplit=1)[1]
                if suffix not in suffix_count:
                    suffix_count[suffix]=0
                suffix_count[suffix]+=1
            except IndexError:
                pass
        for suff,count in suffix_count.items():
            logger.debug(f"Found suffix {suff} {count} times on {tot_files} files")
            if float(count/tot_files) >= threshold:
                ret_suffixes.append("."+suff)
        return ret_suffixes

def get_dataframe(object : object)->pd.DataFrame:
    var_dict=object.__dict__.copy()
    for k,v in var_dict.items():
        var_dict[k]=[v]
    df=pd.DataFrame(var_dict)
    return df  




from typing import Callable,Iterable,Any,Generator

filter_registry:dict[str,Callable[...,bool]]=dict()
sort_registry:dict[str,Callable[...,Iterable]]=dict()
class Filter():
    def __init__(self,fun:Callable[...,bool]):
        self.fun=fun
        
    def run(self,values:Iterable,comparison)->Generator[Any,None,None]:
        for v in values:
            if self.fun(v,comparison):
                yield v
                
class FilterFactory():
    def __init__(self):
        pass
    def create_filter(name:str)->Filter:
        # print(filter_registry)
        if not name in filter_registry:
            return Filter(lambda a: True)
        return Filter(filter_registry[name])
    
def filter(name:str):
    def decorator(fun:Callable[...,bool]):
        filter_registry[name]=fun
        def wrapper():
            fun()
            return
        return wrapper
    return decorator

def sorter_provider(name:str):
    if not name in sort_registry:
        return None
    return sort_registry[name]

def sorter(name:str):
    def decorator(fun:Callable[...,bool]):
        sort_registry[name]=fun
        def wrapper():
            fun()
            return
        return wrapper
    return decorator


from src.utility import filter,sorter
from typing import Any,Iterable
import re
from datetime import datetime
@filter("equal")
def equal(value:Any,comparison:Any)->bool:
    # print(value==comparison)
    if not value:
        return True
    return value==comparison

@filter("like")
def like(value:str,comparison:str)->bool:
    # print(value==comparison)
    if not value:
        return True
    return re.match(f'^{value}.*',comparison) != None

@filter("smaller")
def smaller(value:float,comparison:float)->bool:
    # print(value==comparison)
    if value==None:
        return True
    return value<comparison

@filter("greater")
def greater(value:float,comparison:float)->bool:
    if value==None:
        return True
    # print(value==comparison)
    return value>comparison

@sorter("date")
def sort_dates(values:Iterable,key,reverse:bool=False):
    value_list=list(values)
    value_list.sort(reverse=reverse,key=lambda v: datetime.strptime(key(v),r"%d-%m-%Y"))
    return value_list


import datetime as dt
import sys
import json
import logging
import logging.config
import atexit
from logging.config import ConvertingList, valid_ident
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from atexit import register
LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}




def _resolve_handlers(l):
    if not isinstance(l, ConvertingList):
        return l

    # Indexing the list performs the evaluation.
    return [l[i] for i in range(len(l))]


class QueueListenerHandler(QueueHandler):

    def __init__(self, handlers, respect_handler_level=False, auto_run=True, queue=Queue(-1)):
        super().__init__(queue)
        handlers = _resolve_handlers(handlers)
        self._listener = QueueListener(
            self.queue,
            *handlers,
            respect_handler_level=respect_handler_level)
        if auto_run:
            self.start()
            register(self.stop)


    def start(self):
        self._listener.start()


    def stop(self):
        self._listener.stop()


    def emit(self, record):
        return super().emit(record)

class MyJSONFormatter(logging.Formatter):
    def __init__(
        self,
        *,
        fmt_keys: dict[str, str] | None = None,
    ):
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord):
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.timezone.utc
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: msg_val
            if (msg_val := always_fields.pop(val, None)) is not None
            else getattr(record, val)
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        for key, val in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                message[key] = val

        return message
    
from pathlib import Path
LOG_CONFIGURED=False
def setup_logging(env:str="DEV"):
    global LOG_CONFIGURED
    if not LOG_CONFIGURED:
        parent_dir=Path(__file__).parent
        parent_dir.parent.parent.joinpath("logs").mkdir(exist_ok=True)
        if env=="DEV":
            config_file = parent_dir.joinpath("logging_configs","logs_config_file.json") if sys.version_info.minor >=12 else parent_dir.joinpath("logging_configs","logs_config_file_old.json")
        else:
            config_file = parent_dir.joinpath("logging_configs","logs_config_file_prod.json") if sys.version_info.minor >=12 else parent_dir.joinpath("logging_configs","logs_config_file_old_prod.json")
        with open(config_file) as f_in:
            config = json.load(f_in)
        logging.config.dictConfig(config)
        if sys.version_info.minor >=12 :
            queue_handler = logging.getHandlerByName("queue_handler")
            if queue_handler is not None:
                queue_handler.listener.start()
                atexit.register(queue_handler.listener.stop)
        LOG_CONFIGURED=True





from webbrowser import open
from dash import Dash,html,dcc,page_container,Input,Output,callback,no_update,State,set_props,ctx
from typing import Union,Optional
from datetime import date
from pathlib import Path
from waitress import serve
from concurrent.futures import ProcessPoolExecutor,ThreadPoolExecutor
from repository_miner import RepoMiner
from src.utility import get_dataframe
from src.app.helper import retrieve_contribution_data,parallel_commit_retrievial
import pandas as pd
import dash_bootstrap_components as dbc
import re
from functools import partial
from logging import getLogger
logger=getLogger()

assets_folder=Path(__file__).parent.parent.joinpath("gui","assets")
def start_app(repo_path:Union[str|Path],cicd_test:bool,env:bool):
    path=repo_path if isinstance(repo_path,str) else repo_path.as_posix()
    # print(path)
    pr_name=re.subn(r"_|-"," ",Path(path).name)[0].capitalize()
    app=Dash(name=__name__,title="PVT",assets_folder=assets_folder.as_posix(),external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.BOOTSTRAP,assets_folder.joinpath("css").as_posix()],use_pages=True,pages_folder=Path(__file__).parent.parent.joinpath("gui","pages").as_posix())
    navbar = dbc.NavbarSimple(
        children=[
            dbc.NavItem(dbc.NavLink("Home", href="/")),
            dbc.NavItem(dbc.NavLink("Directory analysis", href="/dir")),
            dbc.Button(id="reload_button",children=[html.I(className="bi bi-arrow-counterclockwise p-1")],className="p1 bg-transparent border-0",value=0),
            dbc.Button(id="open_info",children=[html.I(className="bi bi-list p-1")],className="p1 bg-transparent border-0"),
        ],
        brand=pr_name,
        brand_href="/",
        color="dark",
        dark=True,
        className="mb-2",
        sticky=True,
        fluid=True
    )
    logger.info("Loading necessary info, hang tight")
    contributions,tr_fa=retrieve_contribution_data(path)
    logger.info("You're good to go")
    general_options=dbc.Offcanvas(id="sidebar_info",title="History filters",is_open=False,children=
        [dbc.Stack(
        [
            dbc.RadioItems(id="filter_picker",options=[{"label":"Branch picker","value":"branch_option"},{"label":"Tag picker","value":"tag_option"}],value="branch_option",inline=True,switch=True),
            html.Div([dbc.Label(["Branch Picker"]),dcc.Dropdown(id="branch_picker",searchable=True,clearable=True,placeholder="Branch name")],id="branch_picker_div",className="d-inline"),
            html.Div([dbc.Label(["Tag Picker"]),dcc.Dropdown(id="tag_picker",searchable=True,clearable=True,placeholder="Tag name")],id="tag_picker_div",className="d-inline"),
        ], gap=2,className="p-2")
        ])
    app.layout = html.Div([
        dcc.Store(id="contribution_cache",data=contributions.to_dict("records"),storage_type="session"),
        dcc.Store(id="truck_cache",data=tr_fa,storage_type="session"),
        dcc.Loading(children=[
            dcc.Store(id="branch_cache"),
            dcc.Store(id="commit_df_cache",storage_type="memory"),
            ],fullscreen=True),
        dcc.Store(id="authors_cache"),
        dcc.Store("repo_path",data=path,storage_type="session"),
        navbar,
        general_options,
        dbc.Tooltip("Open filter pickers",target="open_info"),
        dbc.Row([ 
                dbc.Col(
                        children=[
                            dbc.Container([
                            ],fluid=True),
                            page_container
                            ],
                        width=12,align="end"), 
            ],align="start")        
    ])
    
    if not cicd_test:
        if env=="DEV":
            app.run(debug=True,dev_tools_hot_reload=True)
        else:
            open("http://localhost:8080/")
            serve(app.server,host="localhost",port=8080,_quiet=True,threads=100)

@callback(
        Output("commit_df_cache","data"),
        Input("reload_button","n_clicks"),
        State("repo_path","data")
)
def listen_data(_,data):
        rp=RepoMiner(data)
        set_props("branch_picker",{"options":list(( b.name for b in rp.local_branches()))})
        set_props("tag_picker",{"options":list(( b.name for b in rp.get_tags()))})
        commits=parallel_commit_retrievial(rp)
        commit_df=pd.DataFrame(map(lambda c:c.__dict__,commits))
        commit_df["date"]=pd.to_datetime(commit_df["date"])
        commit_df["dow"]=commit_df["date"].dt.day_name()
        commit_df["dow_n"]=commit_df["date"].dt.day_of_week
        commit_df=commit_df.drop("tree_func",axis=1)
        return commit_df.to_dict("records")
        
@callback(
    Output("authors_cache","data"),
    Input("reload_button","n_clicks"),
    State("repo_path","data"),
)
def load_authors(_,data):
    rp=RepoMiner(data)
    authors=pd.DataFrame()
    for author in rp.authors():
        authors=pd.concat([authors,get_dataframe(author)])
    return authors.to_dict("records")


@callback(
        Output("branch_cache","data"),
        Input("branch_picker","value"),
        Input("tag_picker","value"),
        Input("commit_df_cache","data"),
        State("repo_path","data"),
)
def filter_branch_data(v,t,cache,path):
        caller=ctx.triggered_id
        branch=None
        if caller=="branch_picker" and not (not  v or "all" == v ):
            branch = v
        if caller=="tag_picker" and not ( not t or "all" == t):
            branch= t
        if branch:
            df=pd.DataFrame(cache)
            rp=RepoMiner(path)
            b=[c.commit_hash for c in rp.get_branch(branch).traverse_commits()] if caller=="branch_picker" else [c.commit_hash for c in rp.get_tag(branch).traverse_commits()]
            df=df[df["commit_hash"].isin(b)]
            cache = df.to_dict("records")
        cache = cache if cache else no_update
        return cache

@callback(
        Output("branch_picker","disabled"),
        Output("tag_picker","disabled"),
        Output("branch_picker","value"),
        Output("tag_picker","value"),
        Input("filter_picker","value"),
)
def choose_filter(choice):
        output=[False,True,None,None] if choice == "branch_option" else [True,False,None,None]
        return output


from app.cli import main
from pytest import raises,fixture
from sys import argv
from pathlib import Path
workingpath=Path.cwd()
#single line
def test_cli_function():
    argv.clear()
    main(cicd_test=True)

def test_cli():
    main([workingpath.as_posix()],cicd_test=True)

def test_cli_no_dir():
    with raises(SystemExit) as e:
        main([f"{workingpath.joinpath('main.py').as_posix()}"],cicd_test=True)
    assert e.value.code == 1
    
def test_cli_no_git():
    with raises(SystemExit) as e:
        main([f"{workingpath.parent.as_posix()}"],cicd_test=True)
    assert e.value.code ==3


from typing import Generator,Iterable
from src._internal import TreeStructure,File,Folder
from repository_miner import RepoMiner,GitCmdError
from repository_miner.data_typing import Blob,Tree,CommitInfo
from pathlib import Path
import pandas as pd
import truck_factor_gdeluisi.main as tf_calculator
from src.utility.helper import get_dataframe
from os import cpu_count
from math import ceil,floor
from concurrent.futures import ThreadPoolExecutor,ProcessPoolExecutor
from functools import partial
import dash_bootstrap_components as dbc
from dash import html
max_worker = min(32,cpu_count())

def build_tree_structure(miner:RepoMiner,commit_sha:str,path_filter:Iterable=set())->TreeStructure:
    t=miner.tree(commit_sha)
    tree = TreeStructure(hash=t.hash,content=[])
    for o in t.traverse():
        obj=None
        if o.path not in path_filter:
            continue
        if isinstance(o,Blob):
            obj=File(name=o.name,size=o.size,hash_string=o.hash)
        else:
            obj=Folder(name=Path(o.path).name,content=dict(),hash_string=o.hash)
        tree.build(path=o.path,new_obj=obj)
    return tree

def retrieve_SATDs(miner:RepoMiner,satd_highlighters:Iterable[str])->dict[str,dict[int,str]]:
    reg=f'\"[^a-zA-Z0-9_]+\\s*({"|".join(satd_highlighters)}).+\"'
    lines=[]
    try:
        lines=miner.git.grep(["-E","-n","-o","-I",reg])
    except GitCmdError:
        return dict()
    satds:dict[str,dict[int,str]]=dict()
    for line in lines.split('\n'):
        try:
            path,n,satd=line.split(":",2)
            satd=satd.strip()
            res=satds.get(path,{})
            res[n]=satd
            satds[path]=res
        except ValueError:
            print(line)
    return satds

def retrieve_contribution_data(repo_path:str)->tuple[pd.DataFrame,int]:
    contributions=tf_calculator.compute_DOA(tf_calculator.create_contribution_dataframe(repo_path,only_of_files=False))
    tf_contributions=tf_calculator.filter_files_of_interest(contributions)
    tr_fa=tf_calculator.compute_truck_factor_from_contributions(tf_contributions)
    return (contributions,tr_fa)

def parallel_commit_retrievial(rp:RepoMiner)->list[CommitInfo]:
    no_commits=rp.n_commits()
    c_slice=ceil(no_commits/max_worker)
    return_commits=[]
    tasks=[]
    with ThreadPoolExecutor(max_workers=max_worker) as executor:
        for i in range(max_worker):
            tasks.append(executor.submit(rp.retrieve_commit_list,max_count=c_slice,skip=i*c_slice,merges=True))
    for c_list in tasks:
        res=c_list.result()
        if res:
            return_commits.extend(res)
    return return_commits

def create_info_card_columns(texts:dict,icons:dict)->list:
    cols=[]
    fraction=ceil(12/len(texts))
    width=max(fraction,1)
    for title,text in texts.items():
        icon=""
        if title in icons:
            icon = icons[title]
        cols.append(dbc.Col(
                    [
                            dbc.Card([
                                    dbc.CardHeader([icon,title],class_name="fw-bold h5 text-start"),
                                    dbc.CardBody(children=[text],class_name="h5 text-center"),
                            ],class_name="m-2")
                    ]
            ,width=width,align="start"))
    return cols


from pytest import mark,raises,fixture
from pathlib import Path
from _internal.file_parser import fetch_source_files,find_comments_with_locations,find_file_comments_with_locations,find_satd,find_satd_file
from typing import Generator
from utility import logs as log
import logging
log.setup_logging()
logger=logging.getLogger("Parser tester")
workingpath=Path.cwd()
testdata_fetchsource_correct=[
    (workingpath,[".py"],9),
    (workingpath.as_posix(),[".py",".toml"],10),
]
# "Path does not exist"
# "Path is not a directory" .....................
testdata_fetchsource_wrong=[
    (workingpath.joinpath("main.py"),[".py"],"Path is not a directory"),
    (workingpath.joinpath("dont_exist"),[".py"],"Path does not exist"),
    (workingpath.joinpath("main.py"),[".y"],"Path is not a directory"),
]
test_no_file_found=[
    (workingpath.as_posix(),[".y"]),
    (workingpath,[".y"])
    ]

@mark.parametrize("path,extensions,expected",testdata_fetchsource_correct)
def test_fetch_source_files(path,extensions,expected):
    f_list=list(fetch_source_files(path,extensions=extensions))
    assert len(f_list) >= expected
    
@mark.parametrize("path,extensions,expected",testdata_fetchsource_wrong)
def test_fetch_source_files_error(path,extensions,expected):
    with raises((FileNotFoundError,NotADirectoryError)):
        list(fetch_source_files(path,extensions=extensions))
    
@mark.parametrize("path,extensions",test_no_file_found)
def test_no_file_found(path,extensions):
    f_list=list(fetch_source_files(path,extensions=extensions))
    assert len(f_list) >= 0

test_path=workingpath.joinpath("tests")
test_path_content=list(test_path.iterdir())
item=test_path_content.pop().as_posix()
test_path_content.append(item)
@mark.parametrize("path",test_path_content)
def test_find_comments_in_file(path):
    try:
        comments=find_file_comments_with_locations(path)
        if isinstance(path,str):
            path = Path(path)
        logger.debug(f"Found {len(comments)} comments for file {path.as_posix()}")
        logger.debug("Look for comments info in .findings field of this json",extra={"findings":[f"Found comments from {loc[0]} to {loc[1]} in {path.as_posix()}. Comment: {loc[2]}" for loc in comments]})
    except FileNotFoundError as e:
        if isinstance(path,str):
            path = Path(path)
        if path.is_file():
            raise e
        assert str(e) == f"Path {path.as_posix()} is not a file or does not exist"
    else:
        if path.suffix==".txt":
            assert len(comments)==0
        else:
            assert True

def test_find_comments():
    content='import pydriller as git\nimport pydriller.metrics.process.contributors_count as contr\nimport pydriller.metrics.process.history_complexity as history\nimport pydriller.metrics.process.commits_count as comcnt\nfrom .typing import Author\nfrom typing import Optional,Generator,Union\nfrom pathlib import Path\nfrom datetime import datetime\nfrom git import Git,Repo,Blob\nfrom io import BytesIO\nimport re\n\n#TODO implemente a threaded version for optimization\nclass RepoMiner():\n    def __init__(self,repo_path:Union[Path,str]):\n        self.repo_path=repo_path\n        if isinstance(repo_path,Path):\n            self.repo_path=repo_path.as_posix()\n\n    def get_commits_hash(self,since:Optional[datetime]=None,to:Optional[datetime]=None)->Generator[str,None,None]:\n        repo=git.Repository(path_to_repo=self.repo_path,since=since,to=to)\n        return (commit.hash for commit in repo.traverse_commits())\n    \n    def get_commits_between(self,from_commit:str,to_commit:str)->Generator[str,None,None]:\n        repo=git.Repository(path_to_repo=self.repo_path,from_commit=from_commit,to_commit=to_commit)\n        return (commit.hash for commit in repo.traverse_commits())\n    \n    def get_all_authors(self)->set[Author]:\n        repo=git.Repository(path_to_repo=self.repo_path)\n        authors:set[Author]=set((Author(commit.author.email,commit.author.name) for  commit in repo.traverse_commits()))\n        return authors\n\n    def get_author_commits(self,author_name:str)->Generator[str,None,None]:\n        repo=git.Repository(path_to_repo=self.repo_path,only_authors=[author_name])\n        return (commit.hash for commit in repo.traverse_commits())\n\n    def get_file_authors(self,file:str)->set[Author]:\n        repo=git.Repository(path_to_repo=self.repo_path,only_no_merge=True,filepath=file)\n        authors:set[Author]=set((Author(commit.author.email,commit.author.name) for  commit in repo.traverse_commits()))\n        return authors\n    \n    def get_commit(self,commit_hash:str)->git.Commit:\n        repo=git.Repository(path_to_repo=self.repo_path,single=commit_hash)\n        return list(repo.traverse_commits())[0]\n\n    def get_last_modified(self,commit:str):\n        git_repo=git.Git(self.repo_path)\n        return git_repo.get_commits_last_modified_lines(git_repo.get_commit(commit))\n    #TODO include option to use multiple filenames\n    def get_source_code(self,file:Union[str,Path],commit:Optional[str]=None)->list[str]:\n        text=[]\n        file_path=file\n        if isinstance(file,str):\n            file_path=Path(file)\n        git_repo=Repo(self.repo_path)\n        target_commit=git_repo.commit(commit)\n        tree=target_commit.tree\n        #FIXME: improve search method for file\n        for t in tree.traverse():\n            if isinstance(t,Blob) and Path(t.abspath).as_posix()==file_path.as_posix():\n                with BytesIO(t.data_stream.read()) as f:\n           text=re.split(string=f.read().decode(),pattern=r\'\\r\\n|\\n|\\r\')\n        return text\n        # file_path=file\n        # if isinstance(file,Path):\n        #     file_path=file.as_posix()\n        # repo=git.Repository(self.repo_path,single=commit,order="reverse")\n        # list(repo.traverse_commits())[0]\n        # target_commit.\n# def get_diff(repository:str,filepath:str)->dict[str,dict[str, list[tuple[int, str]]]]:\n#     repo=git.Repository(path_to_repo=repository,filepath=filepath,only_no_merge=True,skip_whitespaces=True,order="reverse")\n#     relative_filepath=filepath.removeprefix(repository)[1:].replace("/","\\\\")\n#     diffs:dict[str,dict[str, list[tuple[int, str]]]]=dict()\n#     for commit in repo.traverse_commits():\n#         for f in commit.modified_files:\n#             if f.old_path == relative_filepath or f.new_path == relative_filepath:\n#                 relative_filepath=f.new_path\n#                 if f.old_path != None:\n#                     relative_filepath=f.old_path\n#                 diffs[commit.hash]=f.diff_parsed\n#     return diffs'
    comments=find_comments_with_locations(content,".py")
    logger.debug(f"Found {len(comments)} comments")
    logger.debug("Look for comments info in .findings field of this json",extra={"findings":[f"Found comments from {loc[0]} to {loc[1]}. Comment: {loc[2]}" for loc in comments]})
    assert len(comments)!=0

internal_package=Path.cwd().joinpath("src","_internal")
modules=[item for item in internal_package.iterdir() if item.is_file()]

def test_find_satd():
    path=internal_package.joinpath("file_parser.py")
    try:
        with path.open() as f:
            satds=find_satd(f.read(),path.suffix)
        assert 30 in satds
        assert satds[30] == "TODO could be optimized using multiprocessing"
    except FileNotFoundError as e:
        if path.is_file():
            raise e

@mark.parametrize("path",modules)
def test_find_satd_file(path):
    satds = find_satd_file(filepath=path)
    logger.debug(f"Found these SATDS in file {path.as_posix()}",extra={"satd":satds})


from  dataclasses import dataclass,field
from datetime import date
from time import strptime
from .exceptions import ObjectNotInTreeError,PathNotAvailableError
from typing import Literal,get_args,Iterable,Optional,Union,Generator,Callable
import json
from repository_miner.data_typing import CommitInfoImpl
from src.utility.helper import get_dataframe
import pandas as pd
from pathlib import Path
# ACCEPTED_EXTENSIONS=Literal[".js",".py",".sh",".sql",".c",".cpp",".php",".html",".java",".rb"]
ACCEPTED_EXTENSIONS=Literal[".js",".py",".sh",".sql",".c",".cpp",".php",".html",".java",".rb"]
CONFIG_FOUND=False
config_path=Path(__file__).parent.joinpath("info","ext.json")
if config_path.is_file():
    with config_path.open() as f:
        ACCEPTED_EXTENSIONS:dict[str,str]=json.load(f)
    CONFIG_FOUND=True
    
def check_extension(ext:str,additional_extensions:Iterable[str]=[])->tuple[bool,str]:
    if CONFIG_FOUND:
        if ext not in ACCEPTED_EXTENSIONS:
            return ext in additional_extensions,ext
        return True,ACCEPTED_EXTENSIONS[ext]
    if ext not in get_args(ACCEPTED_EXTENSIONS):
        return ext in additional_extensions,ext
    return True,ext

class RetrieveStrategy():
    def get_source(self,id:str)->list[str]:
        pass
    

# {'commit': '1c85669eb58fc986d43eb7c878e03cb58fb4883d', 'abbreviated_commit': '1c85669', 'tree': 'c6a6edfde2001a68e123c724625faf7599f82371', 'abbreviated_tree': 'c6a6edf', 'parent': 'efe6fba7d02ad06bec603b57f2e5115b7ccd31d8', 'abbreviated_parent': 'efe6fba', 'refs': 'HEAD -> development, origin/development', 'encoding': '', 'subject': 'optimized truck factor function', 'sanitized_subject_line': 'optimized-truck-factor-function', 'body': '', 'commit_notes': '', 'verification_flag': 'N', 'signer': '', 'signer_key': '', 'author': {'name': 'Gerardo De Luisi', 'email': 'deluisigerardo@gmail.com', 'date': 'Sat, 8 Feb 2025 14:21:03 +0100'}, 'commiter': {'name': 'Gerardo De Luisi', 'email': 'deluisigerardo@gmail.com', 'date': 'Sat, 8 Feb 2025 14:21:03 +0100'}}
@dataclass
class SATD():
    satdID: int
    commitID: str
    content: str
    category: str
    file: str
    
@dataclass
class Author():
    email:str
    name:str
    commits_authored:list[str]=field(default_factory=lambda: [])
    def __hash__(self):
        return hash(repr(self.name)+repr(self.email))
    def __eq__(self, value):
        if not isinstance(value,Author):
            raise TypeError(f"Expected value of type <Author>, received {type(value)}")
        return self.name==value.name and self.email==value.email
    def __str__(self):
        return f"Name: {self.name} , Email: {self.email}"
    def __repr__(self):
        return  f"Name: {self.name} , Email: {self.email} , Commits: {self.commits_authored}"

@dataclass
class File():
    name:str
    size:int
    hash_string:str
    def get_source(self,ret_strategy:RetrieveStrategy)->list[str]:
        return ret_strategy.get_source(self.hash_string)
    def __eq__(self, value):
        if not isinstance(value,File):
            raise TypeError(f"Cannot compare type File with {type(value)}")
        return value.hash_string==self.hash_string
    def __hash__(self):
        return self.hash_string.__hash__()
    

@dataclass
class Folder():
    name:str
    content:dict[str,Union[File,'Folder']]
    hash_string:str
    def __eq__(self, value):
        if not isinstance(value,Folder):
            raise TypeError(f"Cannot compare type Folder with {type(value)}")
        return value.hash_string==self.hash_string
    def __hash__(self):
        return self.hash_string.__hash__()
    
    def get_size(self)->int:
        tot_size=0
        for k,v in self.content.items():
            if isinstance(v,File):
                tot_size+=v.size
            else:
                tot_size+=v.get_size()
        return tot_size
    
    def get_dataframe(self):
        size=self.get_size()
        df=pd.DataFrame(dict(size=[size],name=[self.name],hash_string=[self.hash_string]))
        for v in self.content.values():
            df=pd.concat([df,get_dataframe(v)],ignore_index=True).reset_index(drop=True)
        return df

class TreeStructure():
    def __init__(self,hash:str,content:Optional[Iterable[Union[Folder,File]]]=None):
        self.base:Folder=Folder(name="",content=dict(),hash_string=hash)
        self.hash=hash
        if content:
            for c in content:
                self.base.content[c.name]=c
    #Folder->contained_folder/file structure dataframe
    def get_dataframe(self):
        return self.base.get_dataframe()
    
    def get_treemap(self)->dict[str,list[str]]:
        dat_dict:dict[str,list[str]]=dict(parent=[],child=[],name=[],type=[],id=[])
        for path,o in self.walk():
            dat_dict["parent"].append(path if path else "")
            dat_dict["name"].append(o.name)
            dat_dict["id"].append(o.hash_string)
            dat_dict["child"].append(f"{path}/{o.name}" if path else o.name)
            dat_dict["type"].append("folder" if isinstance(o,Folder) else "file")
        return dat_dict
    
    def walk(self,files_only:bool=False,dirs_only:bool=False)->Generator[tuple[str,Union[Folder,File]],None,None]:
        if files_only and dirs_only:
            raise ValueError("Arguments files_only and dirs_only must be mutually exclusive")
        objects=self.base.content.values()
        folders_to_visit:list[tuple[str,Folder]]=[]
        
        end=True
        path=""
        for o in objects:
            if isinstance(o,File):
                if not dirs_only:
                    yield (path,o)
            else:
                folders_to_visit.append((path,o))
                if not files_only:
                    yield (path,o)
                end=False
        while not end:
            path,fold=folders_to_visit.pop()
            path=f"{path}/{fold.name}" if path else fold.name
            for o in fold.content.values():
                if isinstance(o,Folder):
                    folders_to_visit.append((path,o))
                    if not files_only:
                        yield (path,o)
                elif isinstance(o,File) and not dirs_only:
                    yield (path,o)
            end = not folders_to_visit
            
            
    def walk_folder(self,name:str,files_only:bool=False,dirs_only:bool=False)->Generator[Union[Folder,File],None,None]:
        if files_only and dirs_only:
            raise ValueError("Arguments files_only and dirs_only must be mutually exclusive")
        if name not in self.base.content:
            raise ObjectNotInTreeError(f"Object {name} not found among objects {' '.join(self.base.content.keys())}")
        obj=self.base.content[name]
        if isinstance(obj,Folder):
            end=False
            folders_to_visit:list[Folder]=[obj]
            while not end:
                fold=folders_to_visit.pop()
                for o in fold.content.values():
                    if isinstance(o,Folder):
                        folders_to_visit.append(o)
                        if not files_only:
                            yield o
                    elif isinstance(o,File) and not dirs_only:
                        yield o
                end = not folders_to_visit
        else:
            yield obj
    
    def find(self,name:str,type:Optional[Literal["file","folder"]]=None)->Generator[Union[File,Folder],None,None]:
        folder_only=False
        file_only=False
        if type:
            if type=="file":
                file_only=True
            elif type=="folder":
                folder_only=True
            else:
                raise TypeError("The only accepted type values are 'file' and 'folder'")
        for path,o in self.walk(file_only,folder_only):
            if o.name==name:
                yield o
                
    def get(self,path:str)->Union[File,Folder]:
        parts=Path(path).parts
        ret_object=None
        to_explore:list[Folder]=[self.base]

        for p in parts:
            obj=to_explore.pop()
            if p in obj.content:
                if isinstance(obj.content[p],Folder):
                    to_explore.append(obj.content[p])
                ret_object=obj.content[p]
            else:
                raise ObjectNotInTreeError(f"Path {path} is not part of this tree")
        return ret_object
    
    def build(self,path:str,new_obj:Union[File,Folder],mkdir:bool=True):
        if not isinstance(new_obj,(File,Folder)):
            raise TypeError("Object must be of type or subtype of File or Folder")
        if not path:
            raise ValueError("Nonetype or empty string are not amissible paths")
        parts=Path(path).parts
        to_explore:list[Folder]=[self.base]
        num_parts=len(parts)
        last_folder=None
        for i,p in enumerate(parts,1):
            obj=to_explore.pop()
            last_folder=obj
            if p in obj.content:
                if isinstance(obj.content[p],Folder) and not num_parts==i:
                    to_explore.append(obj.content[p])
                else:
                    raise PathNotAvailableError("Either path is not a sequence of folder or path already used")
                
            else:
                if mkdir and not num_parts==i:
                    f=Folder(name=p,content=dict(),hash_string="")
                    obj.content[p]=f
                    to_explore.append(f)
                elif num_parts==i:
                    if not last_folder:
                        self.base.content[new_obj.name]=new_obj
                        return
                    last_folder.content[p]=new_obj
                else:
                    raise ObjectNotInTreeError("Path not reachable")
                
    @staticmethod
    def build_tree(hash:str,path:str,new_obj:Union[File,Folder],mkdir:bool=True)->'TreeStructure':
        tree = TreeStructure(hash=hash)
        if not isinstance(new_obj,(File,Folder)):
            raise TypeError("Object must be of type or subtype of File or Folder")
        if not path:
            raise ValueError("Nonetype or empty string are not amissible paths")
        parts=Path(path).parts
        to_explore:list[Folder]=[tree.base]
        num_parts=len(parts)
        last_folder=None
        for i,p in enumerate(parts,1):
            obj=to_explore.pop()
            last_folder=obj
            if p in obj.content:
                if isinstance(obj.content[p],Folder) and not num_parts==i:
                    to_explore.append(obj.content[p])
                else:
                    raise PathNotAvailableError("Either path is not a sequence of folder or path already used")
                
            else:
                if mkdir and not num_parts==i:
                    f=Folder(name=p,content=dict(),hash_string="")
                    obj.content[p]=f
                    to_explore.append(f)
                elif num_parts==i:
                    if not last_folder:
                        tree.base.content[new_obj.name]=new_obj
                        return
                    last_folder.content[p]=new_obj
                else:
                    raise ObjectNotInTreeError("Path not reachable")
        return tree


import dash
from dash import dcc,callback,Input,Output,no_update,set_props,State,clientside_callback,Patch,ctx,ALL,MATCH
from dash.exceptions import PreventUpdate
import dash.html as html
from datetime import date
import plotly.express as px
import pandas as pd
from dash_ag_grid import AgGrid
from pathlib import Path
from repository_miner import RepoMiner,GitCmdError
from truck_factor_gdeluisi.main import infer_programming_language,resolve_programming_languages
import dash_bootstrap_components as dbc
from io import StringIO
from src.app.helper import build_tree_structure,retrieve_SATDs
import json
import time
from typing import Union
from src._internal.file_parser import DEFAULT_SATD_HIGHLIHGTER
from logging import getLogger
from src.gui import SATDDisplayerAIO
from math import ceil
import re
logger=getLogger("mainpage")
dash.register_page(__name__,"/dir")
items_per_page=5 #TODO make it configurable by user
column_defs_satds=[
        {"field": "fname", 'headerName': 'File',"filter": "agTextColumnFilter"},
        {"field": "line",'headerName': 'Line',"sortable":True},
        {"field": "satd_type", 'headerName': 'Type',"filter": "agTextColumnFilter"},
]
stack=dbc.Stack(id="stack_info",className="p-2 h-75",children=[
        
        dbc.Card(
                id="setd_files_info",
                children=[
                        dbc.CardHeader(id="setd_files_header",children=["SATD discovery",html.I(id="satd_tooltip",className="bi bi-question fw-bold fs-4 ms-3 clickable")]),
                dbc.CardBody(
                        [
                        dbc.Container([
                                AgGrid(
                                        id="satd_table",
                                        columnDefs=column_defs_satds,
                                        columnSize="responsiveSizeToFit",
                                        defaultColDef={"sortable":False,"resizable":True},
                                        dashGridOptions={"pagination": True, "animateRows": False},
                                )
                        ,]
                                ),
                        ])
                        
                ],
        ),
        dbc.Button(id="graph_filtering_collapse_btn",children="Toggle directory filtering"),
        dbc.Collapse([
                dbc.Card([
                dbc.CardHeader([
                        "Graph filtering",
                        html.I(id="dir_filtering_tooltip",className="bi bi-question fw-bold fs-4 ms-3 clickable")
                        ]),
                dbc.CardBody(
                [
                        dbc.Row([
                                dbc.Col(
                                        children=[html.Div([dbc.Label(["Author Picker"]),dcc.Dropdown(id="author_picker",searchable=True,clearable=True,placeholder="Author name")]),],
                                        width=12),
                                        
                                ],className="py-2"),
                        dbc.Row([
                                dbc.Col(
                                        children=[html.Div([dbc.Label(["Degree of Authorship(DOA) threshold picker"]),dcc.Slider(id="doa_picker",min=0,max=1,included=True,step=0.05,value=0.75,marks={"0":"0","1":"1","0.75":"0.75","0.5":"0.5","0.25":"0.25",},tooltip={"placement":"bottom","always_visible":True})]),],
                                        width=12),
                                ],className="py-2"),
                        dbc.Row([
                                dbc.Col(
                                        children=[dbc.Button(id="calculate_doa",children=["Calculate DOAs"],disabled=False)],
                                        width=6),
                                dbc.Col(
                                        children=[dbc.Button(id="reset_doa",children=["Reset"])],
                                        width=6),
                                ],className="py-2",justify="center"),
                ]
                ),]
                ),
        ],"graph_filtering_collapse",is_open=False),
        dbc.Card(
                id="card-file-info",
                children=[
                        dbc.CardHeader(id="file-info-header"),
                        dbc.CardBody(
                        [
                                dcc.Loading([
                                        dcc.Graph(id="file-info-pie")
                                ],overlay_style={"visibility":"visible", "filter": "blur(2px)"}
                                ),
                        ]
                        ),
                ],
                className="invisible"
        ),
        ],gap=2)

layout = dbc.Container([
        dcc.Store("authors_doas",data=dict()),
        dcc.Store("file_cache",data=dict()),
        dcc.Store("file_info_cache",data=dict()),
        dcc.Store("satd_files_cache",data=list()),
        dbc.Tooltip("Click on any cell of a row to look at the SATD of the file at the correspondig line",target="satd_tooltip",trigger="legacy",is_open=False,id="satd_tooltip_info"),
        dbc.Tooltip("Click on the file squares (light blue squares) to obtain informations about their authorship",target="dir_tooltip",trigger="legacy",is_open=False,id="dir_tooltip_info"),
        dbc.Tooltip("Pick an author and a value of DOA to display only the project's submodules on which the author worked",target="dir_filtering_tooltip",trigger="legacy",is_open=False,id="dir_filtering_tooltip_info"),
        dcc.Loading(id="dir_info_loader",display="show",fullscreen=True),
        dbc.Modal([
                        dbc.ModalBody([
                                dbc.Container([
                                html.P(id="line_modal_content",style={"overflow": "hidden","word-wrap":"break-word"})
                        ]),
                    ])
                ],id="line_modal",is_open=False,scrollable=True),
        dbc.Row([
                dbc.Col(
                        [
                        html.H1("Directory Tree Analysis",className="fw-bold h2 px-4 d-inline me-2"),
                        html.I(id="dir_tooltip",className="bi bi-question mb-2 fw-bold d-inline fs-3 clickable")
                        ]
                , width=12,align="end"),
                dbc.Col(
                        [
                        dcc.Loading(id="dir_treemap_loader",
                        children=[
                                dcc.Graph("dir_treemap",className="h-75")
                                ],
                        )
                        ]
                ,width=8,align="start"),
                dbc.Col(
                        [stack],
                        width=4,align="center"
                )
                ]),
                ]
                ,fluid=True)

@callback(
        Output("graph_filtering_collapse","is_open"),
        Input("graph_filtering_collapse_btn","n_clicks"),
        prevent_initial_call=True
)
def open_graph_filtering_collapse(_):
        if _%2>0:
                return True
        return False


@callback(
        # Output("satd_pagination","max_value"),
        Output("satd_table","rowData"),
        Output("satd_files_cache","data"),
        # Input("satd_pagination","active_page"),
        Input("satd_files_cache","data"),
        State("repo_path","data"),
        prevent_intial_call=True
)
def populate_satd_list(cache:dict,path):
        # print("active")
        df=pd.DataFrame()
        if not cache:
                rp=RepoMiner(path)
                highlighters=set(DEFAULT_SATD_HIGHLIHGTER)
                satds=retrieve_SATDs(rp,highlighters)
                cache=satds
        dicts:list[dict]=list()
        for file,satds in cache.items():
                for line,msg in satds.items():
                        msg=msg.strip()
                        t=""
                        pattern=re.compile('|'.join(DEFAULT_SATD_HIGHLIHGTER))
                        t=re.search(pattern,msg).group()
                        dicts.append(dict(fname=file,line=line,satd_type=t))
        df=pd.DataFrame(dicts).to_dict("records")
        # buttons:list[dbc.ListGroupItem]=list()
        # keys=sorted(list(cache.keys()),reverse=True)
        # start_value=(page-1)*items_per_page
        # to_add=keys[start_value:start_value+items_per_page]
        # for key in to_add:
        #         buttons.append(dbc.ListGroupItem(SATDDisplayerAIO(key,cache[key],span_props=dict(className="fw-bold ",style={"cursor":"pointer"}),modal_props={"scrollable":True}).create_comp()))
        return df,cache

@callback(
        Output({"type":"setd_modal","index":MATCH},"is_open"),
        Input({"type":"setd_button","index":MATCH},"n_clicks"),
        prevent_initial_call=True
)
def load_modal(_):
        #load setd modal on file_setd button click
        if _==0:
                return no_update
        return True

@callback(
        Output("file_info_cache","data"),
        Output("dir_treemap","figure"),
        Input("calculate_doa","n_clicks"),
        Input("branch_picker","value"),
        Input("tag_picker","value"),
        Input("contribution_cache","data"),
        State("author_picker","value"),
        State("doa_picker","value"),
        State("repo_path","data"),
)
def populate_treemap(_,b,t,cache,name,doa,data):
        # df=pd.DataFrame(cache)
        if not cache:
                return no_update,no_update
        rp=RepoMiner(data)
        author_doas=None
        tree_dict:dict[str,str]=dict()
        contributions=pd.DataFrame(cache)
        author_doas:pd.DataFrame=contributions.loc[contributions["author"]==name]
        files=contributions["fname"].unique()
        path_filter=set()
        if not author_doas.empty:
                files=author_doas.loc[author_doas["DOA"]>=doa]["fname"].unique()
        path_filter=set(files)
        branch = b if b else t         
        caller=ctx.triggered_id
        tree = build_tree_structure(rp,branch if branch else "HEAD",path_filter)
        for p,o in tree.walk(files_only=True):
                if not p:
                        tree_dict[f"{o.name}"]=o.hash_string
                else:
                        tree_dict[f"{p}/{o.name}"]=o.hash_string
        df=tree.get_treemap()
        df=pd.DataFrame(df)
        fig=px.treemap(data_frame=df,parents=df["parent"],names=df["name"],ids=df["child"],color_discrete_map={'(?)':'lightgrey', 'file':'paleturquoise', 'folder':'crimson'},color=df["type"],custom_data=["id","type"],maxdepth=3,height=800)
        fig.update_layout(
        uniformtext=dict(minsize=10),
        margin = dict(t=50, l=25, r=25, b=25)
        )
        set_props("dir_info_loader",{"display":"auto"})
        return tree_dict,fig

@callback(
        Output("card-file-info","className"),
        Output("file-info-pie","figure"),
        Output("file-info-header","children"),
        Input("dir_treemap","clickData"),
        Input("branch_picker","value"),
        State("contribution_cache","data"),
        State("file-info-header","children"),
        prevent_initial_call=True
)
def populate_file_info(data,_,contributions,children,th=0.7):
        contrs=pd.DataFrame(contributions)
        f_type=data["points"][0]["customdata"][1] if data and data["points"] else "folder"
        if f_type=="folder":
                return "invisible",no_update,no_update
        file=data["points"][0]["id"]
        if children==file:
                return "invisible",no_update,""
        contrs:pd.DataFrame=contrs.loc[contrs["fname"]==file]
        contrs=contrs.loc[contrs["DOA"]>=th]
        top=contrs.groupby("author",as_index=False).sum()
        top=top.sort_values("tot_contributions",ascending=False)
        tot:int=top["tot_contributions"].sum()
        th_percentage=5*tot/100
        top.loc[top['tot_contributions'] < th_percentage, 'author'] = 'Other contributors'
        fig = px.pie(top, values='tot_contributions', names='author',labels={"tot_contributions":"contribution"},title="Contribution distribution")
        return "visible",fig,file

@callback(
        Output("author_picker","value"),
        Output("calculate_doa","n_clicks"),
        Input("reset_doa","n_clicks"),
        )
def reset_options(_):
        if _!=0:
                return None,0
@callback(
        Output("sidebar_info", "is_open"),
        Input("open_info", "n_clicks"),
        [State("sidebar_info", "is_open")],
)
def toggle_offcanvas(n1, is_open):
        if n1:
                return not is_open
        return is_open

@callback(
        Output("author_picker","options"),
        Input("authors_cache","data"),
)
def populate_author_picker(cache):
        authors_df=pd.DataFrame(cache)
        return authors_df["name"].unique().tolist()

@callback(
        Output("calculate_doa","disabled"),
        Input("author_picker","value"),
)
def populate_author_picker(auval):
        return auval==None

@callback(
        Output("line_modal_content","children"),
        Output("line_modal","is_open"),
        Input("satd_table","cellClicked"),
        State("satd_table","rowData"),
        State("satd_files_cache","data"),
)
def populate_satd_modal(cell,row_data,data):
        if not cell:
                raise PreventUpdate()
        row_index=int(cell["rowId"])
        # print(row_index)
        # print(cell)
        row=row_data[row_index]
        fname,line = row["fname"],row["line"]
        # print(fname,line)
        # print(data[fname])
        satd=data[fname][line]
        return satd,True



class ObjectNotInTreeError(Exception):
    def __init__(self, *args):
        super().__init__(*args)

class PathNotAvailableError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


import importlib
import importlib.util
__package_name__="project-visualization-tool-GDeLuisi"
__version__=importlib.metadata.version(__package_name__) if importlib.util.find_spec(__package_name__) else "Version not found"


from dash import Output, Input, State, html, dcc, callback, MATCH,clientside_callback,ALL
import uuid
import dash_bootstrap_components as dbc
from repository_miner.data_typing import Author,CommitInfo
from typing import Iterable
from math import ceil
from src._internal.file_parser import DEFAULT_SATD_HIGHLIHGTER
import re
class AuthorDisplayerAIO(): 
    class ids:
        modal = lambda aio_id: {
            'component': 'AuthorDisplayerAIO',
            'subcomponent': 'modal',
            'aio_id': aio_id
        }
        button = lambda aio_id: {
            'component': 'AuthorDisplayerAIO',
            'subcomponent': 'button',
            'aio_id': aio_id
        }
        pagination =lambda aio_id: {
            'component': 'AuthorDisplayerAIO',
            'subcomponent': 'pagination',
            'aio_id': aio_id
        }
        listgroup =lambda aio_id: {
            'component': 'AuthorDisplayerAIO',
            'subcomponent': 'listgroup',
            'aio_id': aio_id
        }
        store =lambda aio_id: {
            'component': 'AuthorDisplayerAIO',
            'subcomponent': 'store',
            'aio_id': aio_id
        }

    def __init__(
        self,
        author:Author,
        contributions:Iterable[str]=[],
        elements_per_page:int=5,
        text:str="",
        modal_props:dict=None,
        span_props:dict=None,
        div_props:dict=None,
        aio_id:str=None
    ):
        cont=list(sorted(contributions))
        num_pages=ceil(len(cont)/elements_per_page)
        commit_pages=ceil(len(author.commits_authored)/elements_per_page)
        if aio_id is None:
            aio_id = str(uuid.uuid4())
        sp_props= span_props.copy() if span_props else {}
        if "className" not in sp_props:
            sp_props["className"]="clickable"
            
        d_props =div_props.copy() if div_props else {}
        
        m_props = modal_props.copy() if modal_props else {}
        
        self.comp=html.Span([
            dcc.Store(id=self.ids.store(aio_id),data=dict(cont=cont,epg=elements_per_page)),
            dcc.Store(id=self.ids.store(aio_id+"_commit"),data=dict(epg=elements_per_page,cont=author.commits_authored)),
            dbc.Modal([
                dbc.ModalHeader([html.I(className="bi bi-person-circle h3 pe-3"),html.Span(f"{author.name} <{author.email}>",className="fw-bold")]),
                dbc.ModalBody([
                    dbc.Container([

                        html.H6([html.Span(f"Files Authored: ",className="fw-bold") ,str(len(cont))]),
                        html.H6([html.Span(f"Commits Authored: ",className="fw-bold") ,str(len(author.commits_authored))]),
                        html.H6(html.Span("Files authored list: ",className="fw-bold")),

                        dbc.Tabs([
                            dbc.Tab(
                                [
                                    dbc.Pagination(id=self.ids.pagination(aio_id=aio_id),max_value=num_pages,min_value=1,active_page=1,first_last=True, previous_next=True,fully_expanded=False),
                                    dbc.ListGroup(
                                        id=self.ids.listgroup(aio_id=aio_id)
                                    )
                                ] if cont else html.H5("No file authored")
                            ,label="Files Authored"),
                            dbc.Tab(
                                [
                                    dbc.Pagination(id=self.ids.pagination(aio_id=aio_id+"_commit"),max_value=commit_pages,min_value=1,active_page=1,first_last=True, previous_next=True,fully_expanded=False),
                                    dbc.ListGroup(
                                        id=self.ids.listgroup(aio_id=aio_id+"_commit")
                                    )
                                ]
                            ,label="Commits Authored")
                        ]),
                    ]),
                ])
            ],id=self.ids.modal(aio_id),**m_props),
            html.Span(id=self.ids.button(aio_id),children=f"{author.name} {text}",**sp_props)])
            
    def create_comp(self)->html.Span:
        return self.comp 

    clientside_callback(
    """
    function(_,) {
        return {'is_open':true};
    }
    """,
    Output(ids.modal(MATCH), 'is_open'),
    Input(ids.button(MATCH), 'n_clicks'),
    prevent_initial_call=True
    )
    
    clientside_callback(
    """
    function(_, data) {
        const cont=data.cont;
        const epg=data.epg;
        const slicer=epg*(_-1);
        const to_show=cont.slice(slicer,slicer+epg);
        var lis=[];
        to_show.forEach((c)=>{
            lis.push({'type': 'ListGroupItem', 'namespace': 'dash_bootstrap_components', 'props': {'children': c}})
        });
        return lis;
    }
    """,
    Output(ids.listgroup(MATCH), 'children'),
    Input(ids.pagination(MATCH), 'active_page'),
    State(ids.store(MATCH),"data")
    )

class SATDDisplayerAIO(): 
    class satd_ids:
        modal = lambda satd_id: {
            'component': 'SATDDisplayerAIO',
            'subcomponent': 'modal',
            'satd_id': satd_id
        }
        button = lambda satd_id: {
            'component': 'SATDDisplayerAIO',
            'subcomponent': 'button',
            'satd_id': satd_id
        }
        content =lambda satd_id: {
            'component': 'SATDDisplayerAIO',
            'subcomponent': 'content',
            'satd_id': satd_id
        }
        table =lambda satd_id: {
            'component': 'SATDDisplayerAIO',
            'subcomponent': 'table',
            'satd_id': satd_id
        }
        store =lambda satd_id: {
            'component': 'SATDDisplayerAIO',
            'subcomponent': 'store',
            'satd_id': satd_id
        }
        
        table_data=lambda satd_id:{
            'component': 'SATDDisplayerAIO',
            'subcomponent': 'table_data',
            'satd_id': satd_id,
        }

    def __init__(
        self,
        file:str,
        satds:dict[int,str]={},
        text:str="",
        modal_props:dict=None,
        span_props:dict=None,
        div_props:dict=None,
        satd_id:str=None
    ):
        aio_id=satd_id
        if satd_id is None:
            aio_id = str(uuid.uuid4())
        sp_props= span_props.copy() if span_props else {}
        if "style" not in sp_props:
            sp_props["style"]=dict(cursor="pointer")
            
        d_props =div_props.copy() if div_props else {}
        
        m_props = modal_props.copy() if modal_props else {}
        table_header = [html.Thead(html.Tr([html.Th("Line"), html.Th("Type")]))]
        # satd_table_dict:dict[str,list[str]]=dict(type=list(),line=list(),content=list(),placeholder=list())
        rows=list()
        i=0
        for n,c in satds.items():
            c=c.strip()
            t,content="",""
            pattern=re.compile('|'.join(DEFAULT_SATD_HIGHLIHGTER))
            t=re.search(pattern,c).group()
            _,content=re.split(pattern,c,1)
            rows.append(html.Tr([html.Td(str(n),id=self.satd_ids.table_data(aio_id+f"_table_{str(i)}_line"),style={"cursor":"pointer"},className="fw-bold text-info"),
                                html.Td(t,id=self.satd_ids.table_data(aio_id+f"_table_{str(i)}_type")),
                                dbc.Modal([
                                    dbc.ModalBody([
                                        dbc.Container([
                                            html.P(id=self.satd_ids.content(aio_id+f"_table_{str(i)}_line"),style={"overflow": "hidden","word-wrap":"break-word"},children=content)
                                        ]),
                                    ])
                                ],id=self.satd_ids.modal(aio_id+f"_table_{str(i)}_line"),is_open=False,scrollable=True),
                                ]))
            i+=1
        table_body = [html.Tbody(rows,id=self.satd_ids.table(aio_id+"_table"))]
        table = dbc.Table(table_header + table_body, bordered=True,size="sm",)
        self.comp=html.P([
            dbc.Modal([
                dbc.ModalHeader([html.I(className="bi bi-wrench h3 pe-3"),html.Span(f"{file} SATDs",className="fw-bold")]),
                dbc.ModalBody([
                    dbc.Container([
                        html.H6([html.Span("SATDs found:",className="fw-bold") ,str(len(satds.keys()))]),
                        table
                        # dash_table.DataTable(satd_df,[{"name": "line", "id": "line"},{"name": "type", "id": "type"},{"name": "content", "id": "placeholder","presentation":"markdown"}],filter_action="native",sort_action="native", id=self.satd_ids.table(aio_id+"_table"),markdown_options={"html":True} ),
                    ]),
                ])
            ],id=self.satd_ids.modal(aio_id),**m_props),
            html.Span(id=self.satd_ids.button(aio_id),children=file,**sp_props),html.Span(text)
        ])
            
    def create_comp(self)->html.P:
        return self.comp 

    clientside_callback(
    """
    function(_,) {
        return {'is_open':true};
    }
    """,
    Output(satd_ids.modal(MATCH), 'is_open',allow_duplicate=True),
    Input(satd_ids.button(MATCH), 'n_clicks'),
    prevent_initial_call=True
    )
    
    clientside_callback(
    """
    function(_,) {
        return {'is_open':true};
    }
    """,
        Output(satd_ids.modal(MATCH), 'is_open',allow_duplicate=True),
        Input(satd_ids.table(MATCH), 'n_clicks'),
        prevent_initial_call=True
    )
    
    clientside_callback(
    """
    function(_) {
        return {'is_open':true};
    }
    """,
        Output(satd_ids.modal(MATCH), 'is_open'),
        Input(satd_ids.table_data(MATCH), 'n_clicks'),
        prevent_initial_call=True
    )

class CommitDisplayerAIO():
    class ids:
        modal = lambda aio_id: {
            'component': 'CommitDisplayerAIO',
            'subcomponent': 'modal',
            'aio_id': aio_id
        }
        button = lambda aio_id,target: {
            'component': 'CommitDisplayerAIO',
            'subcomponent': 'button',
            'aio_id': aio_id,
            'target':target
        }
        content =lambda aio_id,sub_id: {
            'component': 'CommitDisplayerAIO',
            'subcomponent': 'content',
            'aio_id': aio_id,
            'sub_id':sub_id
        }
        table =lambda aio_id: {
            'component': 'CommitDisplayerAIO',
            'subcomponent': 'table',
            'aio_id': aio_id
        }
        store =lambda aio_id: {
            'component': 'CommitDisplayerAIO',
            'subcomponent': 'store',
            'aio_id': aio_id
        }
        
        table_data=lambda aio_id:{
            'component': 'CommitDisplayerAIO',
            'subcomponent': 'table_data',
            'aio_id': aio_id,
        }

    def __init__(
        self,
        commit:CommitInfo,
        text:str="",
        modal_props:dict=None,
        span_props:dict=None,
        div_props:dict=None,
        id:str=None
    ):
        aio_id=id
        if id is None:
            aio_id = str(uuid.uuid4())
        sp_props= span_props.copy() if span_props else {}
        if "className" not in sp_props:
            sp_props["className"]="clickable"
        d_props =div_props.copy() if div_props else {}
        m_props = modal_props.copy() if modal_props else {}        
        self.comp=html.Span([
            dbc.Modal([
                dbc.ModalHeader([html.I(className="bi bi-git h3 pe-3"),html.Span(f"Commit {commit.abbr_hash}",className="fw-bold")]),
                dbc.ModalBody([
                    dbc.Container([
                        html.P([html.Span("Commit Message: ",className="fw-bold"), commit.subject] ),
                        html.P([html.Span("Commit Author: ",className="fw-bold"), f"{commit.author_name} <{commit.author_email}>"]),
                        html.P([html.Span("Complete hash string: ",className="fw-bold"),commit.commit_hash]),
                        html.P([html.Span("Created at: ",className="fw-bold") ,commit.date.strftime(r"%d-%m-%Y")]),
                        
                    ]),
                ])
            ],id=self.ids.modal(aio_id),size="lg",**m_props),
            html.Span(id=self.ids.button(aio_id,"modal"),children=f"{commit.abbr_hash} {text}",**sp_props)
        ])
    def create_comp(self)->html.Span:
        return self.comp 

    clientside_callback(
    """
    function(_,) {
        return true;
    }
    """,
    Output(ids.modal(MATCH), 'is_open'),
    Input(ids.button(MATCH,"modal"), 'n_clicks'),
    prevent_initial_call=True
    )
